"""
TestCase for SQL.
"""

import pytest
from sqlalchemy import inspect
from sqlalchemy.exc import OperationalError

from ffxiahbot.database import Database

EXPECTED_TABLE_NAMES: set[str] = {"auction_house"}


def test_real_connection(real_db: Database):
    try:
        assert set(inspect(real_db.engine).get_table_names()) == EXPECTED_TABLE_NAMES
    except OperationalError:
        pytest.xfail("OperationalError")


def test_fake_connection(fake_db: Database):
    assert not inspect(fake_db.engine).get_table_names()


def test_populated_fake_db(populated_fake_db: Database):
    assert set(inspect(populated_fake_db.engine).get_table_names()) == EXPECTED_TABLE_NAMES



from datetime import UTC, datetime

import pytest
from sqlalchemy import Null

from ffxiahbot.auction.seller import Seller
from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from ffxiahbot.timeutils import datetime_to_timestamp


@pytest.fixture()
def sell_date() -> datetime:
    return datetime(2000, 2, 2, tzinfo=UTC)


@pytest.fixture()
def sell_timestamp(sell_date: datetime) -> int:
    return datetime_to_timestamp(sell_date)


def test_set_history(populated_fake_db: Database, sell_timestamp) -> None:
    # ensure database has 0 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 0

    # set history for item
    seller = Seller(populated_fake_db, seller=1, seller_name="X", rollback=True, fail=True)
    seller.set_history(itemid=1, stack=False, price=1024, date=sell_timestamp, count=10)

    # ensure database has 10 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.itemid == 1).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.stack == 1).count() == 0
        assert session.query(AuctionHouse).filter(AuctionHouse.stack == 0).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.seller == 1).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.seller_name == "X").count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.date == sell_timestamp).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.price == 1024).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.buyer_name == "X").count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.sale == 1024).count() == 10
        assert session.query(AuctionHouse).filter(AuctionHouse.sell_date == sell_timestamp).count() == 10


def test_sell_item(populated_fake_db: Database, sell_timestamp) -> None:
    # ensure database has 0 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 0

    # sell item
    seller = Seller(populated_fake_db, seller=1, seller_name="X", rollback=True, fail=True)
    seller.sell_item(itemid=1, stack=True, date=sell_timestamp, price=1024, count=5)

    # ensure database has 5 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.itemid == 1).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.stack == 1).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.stack == 0).count() == 0
        assert session.query(AuctionHouse).filter(AuctionHouse.seller == 1).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.seller_name == "X").count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.date == sell_timestamp).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.price == 1024).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.buyer_name == Null).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.sale == 0).count() == 5
        assert session.query(AuctionHouse).filter(AuctionHouse.sell_date == Null).count() == 5



"""
The script will interact with the Auction House of a private Final Fantasy XI server.
"""

import logging
from logging import Handler
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Annotated

import typer
from rich.logging import RichHandler
from typer import Option, Typer

import ffxiahbot.apps.broker
import ffxiahbot.apps.clear
import ffxiahbot.apps.refill
import ffxiahbot.apps.scrub

app = Typer(
    add_completion=False,
    help=__doc__,
    invoke_without_command=True,
    rich_markup_mode="rich",
)


@app.callback()
def setup(
    version: Annotated[bool, Option("--version", help="Show version string.", is_eager=True)] = False,
    silent: Annotated[bool, Option("--silent", help="Only show ERROR messages.")] = False,
    verbose: Annotated[bool, Option("--verbose", help="Also show DEBUG messages.")] = False,
    logfile: Annotated[Path, Option(help="The path to the log file.")] = Path("ahbot.log"),
    no_logfile: Annotated[bool, Option("--disable-logfile", help="Disable logging to a file.")] = False,
) -> None:
    """
    Setup logging.
    """
    if version:
        from ffxiahbot import __version__

        print(f"ffxiahbot v{__version__}")
        typer.Exit(0)

    handlers: list[Handler] = [RichHandler()]

    if not no_logfile:
        file_handler = RotatingFileHandler(logfile, maxBytes=1048576 * 5, backupCount=5)
        file_handler.setFormatter(
            logging.Formatter(
                "[%(asctime)s][%(processName)s][%(threadName)s][%(levelname)-5s]: %(message)s",
                "%Y-%m-%d %H:%M:%S",
            )
        )
        handlers.append(file_handler)

    logging.basicConfig(
        level=(logging.DEBUG if verbose else logging.INFO if not silent else logging.ERROR),
        format="%(message)s",
        handlers=handlers,
    )


app.command("broker")(ffxiahbot.apps.broker.main)
app.command("clear")(ffxiahbot.apps.clear.main)
app.command("refill")(ffxiahbot.apps.refill.main)
app.command("scrub")(ffxiahbot.apps.scrub.main)


if __name__ == "__main__":
    app()



from sqlalchemy import inspect
from sqlalchemy.dialects.mysql import TINYINT
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from ffxiahbot.database import Database
from ffxiahbot.tables.base import Base


def test_compile_tinyint(fake_db: Database) -> None:
    assert Base is not None

    class CustomBase(DeclarativeBase):
        pass

    class CustomTable(CustomBase):
        __tablename__ = "custom_table"

        id: Mapped[int] = mapped_column(
            TINYINT(10, unsigned=True), nullable=False, autoincrement=True, primary_key=True
        )

    CustomBase.metadata.create_all(fake_db.engine)
    assert set(inspect(fake_db.engine).get_table_names()) == {"custom_table"}



from collections.abc import Callable
from typing import Any

from sqlalchemy import DDLElement
from sqlalchemy.dialects.mysql import TINYINT
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.orm import DeclarativeBase


class Base(DeclarativeBase):
    pass


@compiles(TINYINT, "sqlite")
def compile_tinyint(element: DDLElement, compiler: Callable[[Any], str], **kwargs: Any) -> str:
    """
    Compile TINYINT for SQLite test database.
    """
    return "INTEGER"



from dataclasses import dataclass

from ffxiahbot.auction.worker import Worker
from ffxiahbot.logutils import capture
from ffxiahbot.tables.auctionhouse import AuctionHouse


@dataclass(frozen=True)
class Seller(Worker):
    """
    Auction House seller.

    Args:
        db: The database object.
        seller: The auction house seller id.
        seller_name: The auction house seller name.
    """

    #: The auction house seller id.
    seller: int
    #: The auction house seller name.
    seller_name: str

    def set_history(self, itemid: int, stack: int | bool, price: int, date: int, count: int = 1) -> None:
        """
        Set the history of a particular item.

        Args:
            itemid: The item number.
            stack: The stack size.
            price: The price.
            date: The timestamp.
            count: The number of rows.
        """
        with capture(fail=self.fail):
            itemid = AuctionHouse.validate_itemid(itemid)
            stack = AuctionHouse.validate_stack(stack)
            price = AuctionHouse.validate_price(price)
            date = AuctionHouse.validate_date(date)

            # add row
            with self.scoped_session() as session:
                # add the item multiple times
                for _i in range(count):
                    row = AuctionHouse(
                        itemid=itemid,
                        stack=stack,
                        seller=self.seller,
                        seller_name=self.seller_name,
                        date=date,
                        price=price,
                        buyer_name=self.seller_name,
                        sale=price,
                        sell_date=date,
                    )
                    session.add(row)

    def sell_item(self, itemid: int, stack: int, date: int, price: int, count: int) -> None:
        """
        Put up a particular item for sale.

        Args:
            itemid: The item number.
            stack: The stack size.
            date: The timestamp.
            price: The price.
            count: The number of rows.
        """
        with capture(fail=self.fail):
            itemid = AuctionHouse.validate_itemid(itemid)
            stack = AuctionHouse.validate_stack(stack)
            price = AuctionHouse.validate_price(price)
            date = AuctionHouse.validate_date(date)

            # add row
            with self.scoped_session() as session:
                # add the item multiple times
                for _i in range(count):
                    row = AuctionHouse(
                        itemid=itemid,
                        stack=stack,
                        seller=self.seller,
                        seller_name=self.seller_name,
                        date=date,
                        price=price,
                    )
                    session.add(row)



"""
Refill the auction house.
"""

from pathlib import Path
from typing import Annotated

from typer import Option, confirm

from ffxiahbot.common import OptionalPath, OptionalPathList
from ffxiahbot.config import Config
from ffxiahbot.database import Database
from ffxiahbot.logutils import logger
from ffxiahbot.tables.base import Base


def main(
    cfg_path: Annotated[Path, Option("--config", help="Config file path.")] = Path("config.yaml"),
    inp_csvs: Annotated[
        OptionalPathList, Option("--inp-csv", help="Input CSV file path.", show_default="items.csv")
    ] = None,
    no_prompt: Annotated[bool, Option("--no-prompt", help="Do not ask for confirmation.")] = False,
    use_sqlite_db: Annotated[OptionalPath, Option(help="Use a test SQLite database instead of the real one?")] = None,
) -> None:
    """
    Refill the auction house with the items defined in the CSV file.
    """
    from ffxiahbot.auction.manager import Manager
    from ffxiahbot.itemlist import ItemList

    if inp_csvs is None:
        inp_csvs = [Path("items.csv")]

    config: Config = Config.from_yaml(cfg_path)
    logger.info("%s", config.model_dump_json(indent=2))

    # create auction house manager
    if use_sqlite_db:
        manager = Manager.from_db(
            db=Database.sqlite(database=str(use_sqlite_db)),
            name=config.name,
            fail=config.fail,
        )
        Base.metadata.create_all(manager.db.engine)
    else:
        manager = Manager.create_database_and_manager(
            hostname=config.hostname,
            database=config.database,
            username=config.username,
            password=config.password,
            port=config.port,
            name=config.name,
            fail=config.fail,
        )

    # load data
    item_list = ItemList.from_csv(*inp_csvs)

    if no_prompt or confirm("Restock all items?", abort=True, show_default=True):
        logger.info("restocking...")
        manager.restock_items(item_list=item_list, use_selling_rates=config.use_selling_rates)
        logger.info("exit after restock")



from __future__ import annotations

from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from tests.cookbook import AuctionHouseRowBuilder


def test_create_ah_row(populated_fake_db: Database) -> None:
    # check that the table is empty
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 0

    # add a row to the table
    random_items = [AuctionHouseRowBuilder() for _ in range(10)]
    for item in random_items:
        item.add_item_for_sale(populated_fake_db)

    # check that a single row now exists with the expected values
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == len(random_items)
        for i, random_item in enumerate(random_items):
            random_item.validate_query_result(session.query(AuctionHouse).get(i + 1))

    # clear the table
    with populated_fake_db.scoped_session() as session:
        session.query(AuctionHouse).delete()

    # check that the table is empty
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 0



from ffxiahbot.auction.worker import Worker
from ffxiahbot.database import Database


def test_create_worker(populated_fake_db: Database) -> None:
    worker = Worker(populated_fake_db, rollback=True, fail=True)
    assert worker.fail is True
    assert worker.rollback is True
    assert worker.db is populated_fake_db

    with worker.scoped_session():
        pass



import logging
import os
import random
from pathlib import Path

import pandas as pd
import pytest

from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from ffxiahbot.tables.base import Base


@pytest.fixture(scope="function")
def db_path() -> Path:
    return Path(__file__).parent.joinpath("test.db")


@pytest.fixture(scope="function")
def fake_db(db_path: Path) -> Database:
    assert AuctionHouse is not None
    logging.getLogger("sqlalchemy.pool").setLevel(logging.DEBUG)
    logging.getLogger("sqlalchemy.engine").setLevel(logging.INFO)
    yield Database.sqlite(database=str(db_path))
    db_path.unlink(missing_ok=True)


@pytest.fixture(scope="function")
def populated_fake_db(fake_db: Database) -> Database:
    Base.metadata.create_all(fake_db.engine)
    return fake_db


@pytest.fixture(scope="function")
def real_db() -> Database:
    return Database.pymysql(
        hostname=os.environ.get("FFXIAHBOT_TEST_REAL_DB_HOSTNAME", "127.0.0.1"),
        database=os.environ.get("FFXIAHBOT_TEST_REAL_DB_DATABASE", "xidb"),
        username=os.environ.get("FFXIAHBOT_TEST_REAL_DB_USERNAME", "xi"),
        password=os.environ.get("FFXIAHBOT_TEST_REAL_DB_PASSWORD", "password"),
        port=3306,
    )


@pytest.fixture(scope="session", autouse=True)
def set_random_seed():
    random.seed(12345)


@pytest.fixture(scope="session", autouse=True)
def set_pandas_options():
    pd.set_option("display.max_columns", None)
    pd.set_option("display.max_rows", None)
    pd.set_option("display.width", None)
    pd.set_option("display.max_colwidth", None)



from __future__ import annotations

import collections
import functools
from collections.abc import Generator

from pydantic import AliasChoices, BaseModel, ConfigDict, Field

fmt = collections.OrderedDict()
fmt["itemid"] = "{:>8}"
fmt["name"] = "{:>24}"
fmt["sell_single"] = "{:>11}"
fmt["buy_single"] = "{:>11}"
fmt["price_single"] = "{:>16}"
fmt["stock_single"] = "{:>12}"
fmt["buy_rate_single"] = "{:>16}"
fmt["sell_rate_single"] = "{:>16}"
fmt["sell_stacks"] = "{:>11}"
fmt["buy_stacks"] = "{:>11}"
fmt["price_stacks"] = "{:>16}"
fmt["stock_stacks"] = "{:>12}"
fmt["buy_rate_stacks"] = "{:>16}"
fmt["sell_rate_stacks"] = "{:>16}"


def item_csv_title_str() -> str:
    objs = []
    for key in fmt:
        objs.append(fmt[key].format(key))

    return ", ".join(objs) + "\n"


def item_csv_value_str(item: Item) -> str:
    objs = []
    for key in fmt:
        value = getattr(item, key, None)
        objs.append(fmt[key].format(value))

    return ", ".join(objs) + "\n"


_template = """
[Item]
    id         = {self.itemid}
    addr       = {addr}
    name       = {self.name}

[Item.Single]
    buy        = {self.buy_single}
    sell       = {self.sell_single}
    price      = {self.price_single}
    stock      = {self.stock_single}
    buy_rate   = {self.buy_rate_single}
    sell_rate  = {self.sell_rate_single}

[Item.Stacks]
    buy        = {self.buy_stacks}
    sell       = {self.sell_stacks}
    price      = {self.price_stacks}
    stock      = {self.stock_stacks}
    buy_rate   = {self.buy_rate_stacks}
    sell_rate  = {self.sell_rate_stacks}
"""[:-1]


class Item(BaseModel):
    """
    Item properties.

    Args:
        itemid: A unique item id.
        name: The item's name.
        sell_single: sell singles?
        buy_single: buy singles?
        price_single: price (>= 1) for singles.
        stock_single: restock count (>= 0) for singles.
        sell_stacks: sell stacks?
        buy_stacks: buy stacks?
        price_stacks: price (>= 1) for stacks.
        stock_stacks: restock count (>= 0) for stacks.
        buy_rate_single: buy rate (0.0 <= rate <= 1.0) for singles.
        sell_rate_single: sell rate (0.0 <= rate <= 1.0) for singles.
        sell_rate_stacks: sell rate (0.0 <= rate <= 1.0) for stacks.
        buy_rate_stacks: buy rate (0.0 <= rate <= 1.0) for stacks.
    """

    model_config = ConfigDict(extra="forbid")

    #: A unique item id.
    itemid: int = Field(ge=0, validation_alias=AliasChoices("id", "itemid"))
    #: The item's name.
    name: str = "?"
    #: Sell singles?
    sell_single: bool = True
    #: Sell stacks?
    sell_stacks: bool = True
    #: Buy singles?
    buy_single: bool = True
    #: Buy stacks?
    buy_stacks: bool = True
    #: Price (>= 1) for singles.
    price_single: int = Field(default=1, gt=0)
    #: Price (>= 1) for stacks.
    price_stacks: int = Field(default=1, gt=0)
    #: Restock count (>= 0) for singles.
    stock_single: int = Field(default=0, ge=0)
    #: Restock count (>= 0) for stacks.
    stock_stacks: int = Field(default=0, ge=0)
    #: Buy rate (0.0 <= rate <= 1.0) for singles.
    buy_rate_single: float = Field(default=1.0, ge=0.0, le=1.0, validation_alias=AliasChoices("buy_rate_single"))
    #: Sell rate (0.0 <= rate <= 1.0) for singles.
    sell_rate_single: float = Field(
        default=1.0, ge=0.0, le=1.0, validation_alias=AliasChoices("sell_rate_single", "rate_single")
    )
    #: Buy rate (0.0 <= rate <= 1.0) for stacks.
    buy_rate_stacks: float = Field(default=1.0, ge=0.0, le=1.0, validation_alias=AliasChoices("buy_rate_stacks"))
    #: Sell rate (0.0 <= rate <= 1.0) for stacks.
    sell_rate_stacks: float = Field(
        default=1.0, ge=0.0, le=1.0, validation_alias=AliasChoices("sell_rate_stacks", "rate_stacks")
    )

    @classmethod
    def aliases(cls) -> Generator[str]:
        yield from cls.model_fields.keys()

    def __str__(self) -> str:
        return _template.format(self=self, addr=hex(id(self)))


@functools.lru_cache(maxsize=1)
def allowed_item_keys() -> set[str]:
    def _() -> Generator[str]:
        for key, field_info in Item.model_fields.items():
            yield key
            if hasattr(field_info, "validation_alias") and isinstance(field_info.validation_alias, AliasChoices):
                for alias in field_info.validation_alias.choices:
                    if isinstance(alias, str):
                        yield alias

    return set(_())



import pytest

from ffxiahbot.auction.browser import Browser
from ffxiahbot.database import Database
from tests.cookbook import RANDOMIZE as R
from tests.cookbook import AuctionHouseRowBuilder as AHR
from tests.cookbook import setup_ah_transactions


def test_browser_count_empty(populated_fake_db: Database) -> None:
    browser = Browser(populated_fake_db, rollback=True, fail=True)
    assert browser.count() == 0


def test_browser_count_one_item(populated_fake_db: Database) -> None:
    browser = Browser(populated_fake_db, rollback=True, fail=True)
    AHR().add_item_for_sale(populated_fake_db)
    assert browser.count() == 1


@pytest.mark.parametrize(
    "transactions",
    [
        pytest.param(
            [
                *AHR.many(count=1, seller=1, seller_name="A", itemid=1, stack=1, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=2, seller=1, seller_name="A", itemid=2, stack=1, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=3, seller=1, seller_name="A", itemid=1, stack=0, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=4, seller=1, seller_name="A", itemid=2, stack=0, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=5, seller=1, seller_name="A", itemid=1, stack=1),
                *AHR.many(count=6, seller=1, seller_name="A", itemid=2, stack=1),
                *AHR.many(count=7, seller=1, seller_name="A", itemid=1, stack=0),
                *AHR.many(count=8, seller=1, seller_name="A", itemid=2, stack=0),
            ],
            id="one seller",
        ),
        pytest.param(
            [
                *AHR.many(count=1, seller=1, seller_name="A", itemid=1, stack=1, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=2, seller=1, seller_name="A", itemid=2, stack=1, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=3, seller=1, seller_name="A", itemid=1, stack=0, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=4, seller=1, seller_name="A", itemid=2, stack=0, buyer_name="B", sell_date=R, sale=R),
                *AHR.many(count=5, seller=1, seller_name="A", itemid=1, stack=1),
                *AHR.many(count=6, seller=1, seller_name="A", itemid=2, stack=1),
                *AHR.many(count=7, seller=1, seller_name="A", itemid=1, stack=0),
                *AHR.many(count=8, seller=1, seller_name="A", itemid=2, stack=0),
                *AHR.many(count=8, seller=2, seller_name="B", itemid=1, stack=1, buyer_name="A", sell_date=R, sale=R),
                *AHR.many(count=7, seller=2, seller_name="B", itemid=2, stack=1, buyer_name="A", sell_date=R, sale=R),
                *AHR.many(count=6, seller=2, seller_name="B", itemid=1, stack=0, buyer_name="A", sell_date=R, sale=R),
                *AHR.many(count=5, seller=2, seller_name="B", itemid=2, stack=0, buyer_name="A", sell_date=R, sale=R),
                *AHR.many(count=4, seller=2, seller_name="B", itemid=1, stack=1),
                *AHR.many(count=3, seller=2, seller_name="B", itemid=2, stack=1),
                *AHR.many(count=2, seller=2, seller_name="B", itemid=1, stack=0),
                *AHR.many(count=1, seller=2, seller_name="B", itemid=2, stack=0),
            ],
            id="two sellers",
        ),
    ],
)
def test_browser_get_stock_and_price(populated_fake_db: Database, transactions: tuple[AHR]) -> None:
    builder = setup_ah_transactions(populated_fake_db, *transactions)
    browser = Browser(populated_fake_db, rollback=True, fail=True)

    # validate totals being sold
    for i in builder.itemid.unique():
        assert browser.get_stock(i, stack=True, seller=None) == len(
            builder.query(f"itemid == {i} & stack == 1 & sale == 0")
        )
        assert browser.get_stock(i, stack=False, seller=None) == len(
            builder.query(f"itemid == {i} & stack == 0 & sale == 0")
        )

    # validate totals being sold for seller
    for seller in builder.seller.unique():
        for i in builder.itemid.unique():
            assert browser.get_stock(i, stack=True, seller=seller) == len(
                builder.query(f"itemid == {i} & stack == 1 & sale == 0 & seller == {seller}")
            )
            assert browser.get_stock(i, stack=False, seller=seller) == len(
                builder.query(f"itemid == {i} & stack == 0 & sale == 0 & seller == {seller}")
            )

    # validate min price for item up for sale
    for i in builder.itemid.unique():
        assert (
            browser.get_price(i, stack=True, seller=None)
            == builder.query(f"itemid == {i} & stack == 1 & sale > 0").sale.min()
        )
        assert (
            browser.get_price(i, stack=False, seller=None)
            == builder.query(f"itemid == {i} & stack == 0 & sale > 0").sale.min()
        )

    # validate min price for item up for sale by seller
    for seller in builder.seller.unique():
        for i in builder.itemid.unique():
            assert (
                browser.get_price(i, stack=True, seller=seller)
                == builder.query(f"itemid == {i} & stack == 1 & sale > 0 & seller == {seller}").sale.min()
            )
            assert (
                browser.get_price(i, stack=False, seller=seller)
                == builder.query(f"itemid == {i} & stack == 0 & sale > 0 & seller == {seller}").sale.min()
            )



import datetime
import logging
import os
import re
import shutil
from collections.abc import Iterator
from pathlib import Path
from typing import Any, Optional

OptionalPath = Optional[Path]
OptionalIntList = Optional[list[int]]
OptionalStrList = Optional[list[str]]
OptionalPathList = Optional[list[Path]]


def create_path(
    *args: str,
    absolute: bool = True,
    dt_fmt: str = "%Y_%m_%d_%H_%M_%S",
    dt: datetime.datetime | None = None,
    **kwargs: Any,
) -> bytes | str:
    """
    Construct a path.  You can access dt or datetime as objects.

    :param args: path components passed to :py:func:`os.path.join`
    :param absolute: make path absolute
    :param dt: datetime object
    :param dt_fmt: date format
    :param kwargs: variables passed to :py:func:`str.format`

    :type absolute: bool
    :type dt: :py:class:`datetime.datetime`
    :type dt_fmt: str
    """
    if not dt:
        dt = datetime.datetime.now()

    dt_str = dt.strftime(dt_fmt)
    _kwargs = {"dt": dt, "datetime": dt, "date": dt_str}
    _kwargs.update(**kwargs)

    path = os.path.expanduser(os.path.join(*args).format(**_kwargs))

    if absolute:
        return os.path.abspath(path)

    return path


def backup(path: str | Path, copy: bool = False) -> Path | None:
    """
    Create backup file name.

    :param path: path of file
    :param copy: perform copy file
    """
    old_path = os.path.abspath(os.path.expanduser(path))

    # nothing to back up
    if not os.path.exists(old_path):
        return None

    if not os.path.isfile(old_path):
        raise RuntimeError(f"can only backup files: {old_path}")

    # extract file info
    dname, bname = os.path.dirname(old_path), os.path.basename(old_path)
    root, dirs, files = next(os.walk(dname))

    # look for existing backups
    regex = re.compile(rf"^{bname}(\.(\d+))?$")
    found = []
    for f in files:
        match = regex.match(f)
        if match:
            try:
                found.append(int(match.group(2)))
            except TypeError:
                found.append(0)

    # should have at least found 1 if the file exists
    if not found:
        raise RuntimeError(f"can not backup file: {old_path}")

    # create backup name
    new_path = os.path.join(dname, f"{bname}.{max(found) + 1}")

    # validate old_path
    if os.path.exists(new_path) or new_path == old_path:
        raise RuntimeError(f"can not backup file: {old_path}")

    # copy the file
    if copy:
        logging.debug("backup (old): %s", old_path)
        logging.debug("backup (new): %s", new_path)
        shutil.copy(old_path, new_path)

    return Path(new_path)


def find_files(
    top: str, regex: str | re.Pattern[str] = r".*", r: bool = False, ignorecase: bool = True, **kwargs: Any
) -> Iterator[str]:
    """
    Search for files that match pattern.

    :param ignorecase: ignore case in regular expression
    :param top: top level directory
    :param regex: pattern to match
    :param r: recursive search
    """
    regex = (
        regex
        if isinstance(regex, re.Pattern)
        else (re.compile(regex, re.IGNORECASE) if ignorecase else re.compile(regex))
    )

    if r:
        for root, _dirs, files in os.walk(top, **kwargs):
            for f in files:
                match = regex.match(f)
                if match:
                    yield os.path.join(root, f)
    else:
        root, dirs, files = next(os.walk(top, **kwargs))
        for f in files:
            match = regex.match(f)
            if match:
                yield os.path.join(root, f)



import pytest

from ffxiahbot.auction.manager import Manager
from ffxiahbot.database import Database
from ffxiahbot.item import Item
from ffxiahbot.itemlist import ItemList
from ffxiahbot.tables.auctionhouse import AuctionHouse
from tests.cookbook import RANDOMIZE as R
from tests.cookbook import AuctionHouseRowBuilder as AHR
from tests.cookbook import setup_ah_transactions


@pytest.fixture()
def item_list() -> ItemList:
    # fmt: off
    return ItemList(
        items={
            1: Item(
                itemid=1, name="Fake Item A",
                sell_single=True, buy_single=True, price_single=100, stock_single=2, rate_single=1.0,
                sell_stacks=True, buy_stacks=True, price_stacks=200, stock_stacks=4, rate_stacks=1.0,
            ),
            2: Item(
                itemid=2, name="Fake Item B",
                sell_single=True, buy_single=True, price_single=200, stock_single=6, rate_single=1.0,
                sell_stacks=True, buy_stacks=True, price_stacks=400, stock_stacks=8, rate_stacks=1.0,
            ),
            4: Item(
                itemid=2, name="Fake Item C",
                sell_single=True, buy_single=True, price_single=600, stock_single=0, rate_single=1.0,
                sell_stacks=True, buy_stacks=True, price_stacks=800, stock_stacks=0, rate_stacks=1.0,
            ),
        }
    )
    # fmt: on


@pytest.mark.parametrize(
    "transactions,expected_blacklist,expect_updates",
    [
        pytest.param(
            [
                # 1: already bought (stack)
                AHR(price=100, seller=1, seller_name="A", itemid=1, stack=1, buyer_name="X", sell_date=R, sale=R),
                # 2: already bought (single)
                AHR(price=100, seller=1, seller_name="A", itemid=1, stack=0, buyer_name="X", sell_date=R, sale=R),
                # 3: item too expensive (stacks)
                AHR(price=300, seller=1, seller_name="A", itemid=1, stack=1),
                # 4: item too expensive (singles)
                AHR(price=300, seller=1, seller_name="A", itemid=1, stack=0),
                # 3: item not in database (stacks)
                AHR(price=100, seller=1, seller_name="A", itemid=3, stack=1),
                # 4: item not in database (singles)
                AHR(price=100, seller=1, seller_name="A", itemid=3, stack=0),
            ],
            {3, 4, 5, 6},
            {},
            id="Nothing to buy",
        ),
        pytest.param(
            [
                # 1: already bought (stack)
                AHR(price=100, seller=1, seller_name="A", itemid=1, stack=1, buyer_name="X", sell_date=R, sale=R),
                # 2: already bought (single)
                AHR(price=100, seller=1, seller_name="A", itemid=1, stack=0, buyer_name="X", sell_date=R, sale=R),
                # 3: item just right (stacks)
                AHR(price=200, seller=1, seller_name="A", itemid=1, stack=1),
                # 4: item just right (singles)
                AHR(price=100, seller=1, seller_name="A", itemid=1, stack=0),
                # 3: item not in database (stacks)
                AHR(price=100, seller=1, seller_name="A", itemid=3, stack=1),
                # 4: item not in database (singles)
                AHR(price=100, seller=1, seller_name="A", itemid=3, stack=0),
            ],
            {5, 6},
            {3, 4},
            id="Many things to buy",
        ),
    ],
)
def test_buy_items(
    populated_fake_db: Database,
    transactions: tuple[AHR, ...],
    expected_blacklist: set[int],
    expect_updates: set[int],
    item_list: ItemList,
) -> None:
    setup_ah_transactions(populated_fake_db, *transactions)
    manager = Manager.from_db(populated_fake_db, name="X", rollback=True, fail=True)

    # perform buy loop multiple times
    for _ in range(3):
        manager.buy_items(item_list, use_buying_rates=False)
        assert manager.blacklist == expected_blacklist

    # validate database after buy loop
    for i in range(len(transactions)):
        ahr = transactions[i]
        with populated_fake_db.scoped_session() as session:
            row = session.query(AuctionHouse).filter(AuctionHouse.id == i + 1).one()
            if i + 1 in expect_updates:
                assert row.sell_date != 0
                assert row.sale == ahr.price
                assert row.buyer_name == "X"
            else:
                ahr.validate_query_result(row, id=i + 1)


@pytest.mark.parametrize(
    "transactions,expected_history,expected_stock",
    [
        pytest.param(
            [
                *AHR.many(count=10, seller=1, seller_name="A", itemid=4, stack=1),
                *AHR.many(count=10, seller=1, seller_name="A", itemid=4, stack=0),
            ],
            {1: [100, 200], 2: [200, 400]},
            {1: [2, 4], 2: [6, 8], 3: [0, 0], 4: [10, 10]},
            id="Simple restock",
        ),
    ],
)
def test_restock_items(
    populated_fake_db: Database,
    transactions: tuple[AHR, ...],
    expected_history: dict[int, tuple[int, int]],
    expected_stock: dict[int, tuple[int, int]],
    item_list: ItemList,
) -> None:
    if transactions:
        setup_ah_transactions(populated_fake_db, *transactions)

    manager = Manager.from_db(populated_fake_db, name="X", rollback=True, fail=True)

    # perform restock loop multiple times
    for _ in range(3):
        manager.restock_items(item_list, use_selling_rates=False)

    # validate the historical prices after restock loop
    for itemid, (singles_price, stacks_price) in expected_history.items():
        with populated_fake_db.scoped_session() as session:
            stacks_row = (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 1, AuctionHouse.sale > 0)
                .one()
            )
            singles_row = (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 0, AuctionHouse.sale > 0)
                .one()
            )
            assert stacks_row.sale == stacks_price
            assert singles_row.sale == singles_price

    # validate the restocked items after restock loop
    for itemid, (singles_stock, stacks_stock) in expected_stock.items():
        with populated_fake_db.scoped_session() as session:
            stacks_rows = session.query(AuctionHouse).filter(
                AuctionHouse.itemid == itemid, AuctionHouse.stack == 1, AuctionHouse.sale == 0
            )
            singles_rows = session.query(AuctionHouse).filter(
                AuctionHouse.itemid == itemid, AuctionHouse.stack == 0, AuctionHouse.sale == 0
            )
            assert stacks_rows.count() == stacks_stock
            assert singles_rows.count() == singles_stock



"""
The configuration for the bot.
"""

from __future__ import annotations

from pathlib import Path

import yaml
from pydantic import BaseModel, ConfigDict, Field, SecretStr

from ffxiahbot.logutils import logger


# noinspection PyArgumentList
class Config(BaseModel):
    """
    The configuration for the bot.
    """

    model_config = ConfigDict(extra="forbid")

    # Bot
    name: str = Field(default="M.H.M.U.")  # Bot name
    tick: int = Field(default=30)  # Tick interval (seconds)
    restock: int = Field(default=3600)  # Restock interval (seconds)
    use_buying_rates: bool = Field(default=False)  # Only buy items a fraction of the time?
    use_selling_rates: bool = Field(default=False)  # Only sell items a fraction of the time?

    # Database
    hostname: str = Field(default="127.0.0.1")  # SQL address
    database: str = Field(default="xidb")  # SQL database
    username: str = Field(default="xi")  # SQL username
    password: SecretStr | str = Field(default=SecretStr("password"))  # SQL password
    port: int = Field(default=3306)  # SQL port
    fail: bool = Field(default=False)  # Fail on SQL errors?

    @classmethod
    def from_yaml(cls, cfg_path: Path | None = None) -> Config:
        """
        Load the configuration from a file.

        Args:
            cfg_path: The path to the configuration file.

        Returns:
            The configuration instance.
        """
        if cfg_path is None:
            return cls()

        with cfg_path.open("r") as stream:
            data = yaml.safe_load(stream)
            for key in data:
                if key.lower() in DEPRECATED:
                    logger.error(f"Ignoring deprecated config key: {key}")

            return cls(**{k: v for k, v in data.items() if k.lower() not in DEPRECATED})


DEPRECATED = {
    "data": "use --inp-csv",
    "overwrite": "use --overwrite",
    "backup": "use --backup",
    "stub": "use --out-csv",
    "server": "use --server",
    "threads": "use --threads",
    "stock_stacks": "use --default-stock-stack",
    "stock_single": "use --default-stock-single",
    "itemids": "use --item-ids",
    "urls": "use --urls",
    "verbose": "use --verbose",
    "silent": "use --silent",
}



from pathlib import Path

import pytest
from pytest import MonkeyPatch

import ffxiahbot.apps.refill
from ffxiahbot.auction.manager import Manager
from ffxiahbot.config import Config
from ffxiahbot.database import Database
from ffxiahbot.item import Item
from ffxiahbot.itemlist import ItemList
from ffxiahbot.tables.auctionhouse import AuctionHouse


@pytest.fixture
def config() -> Config:
    return Config()


@pytest.fixture
def item_list() -> ItemList:
    return ItemList(
        items={
            1: Item(
                itemid=1,
                name="Fake Item A",
                sell_single=True,
                buy_single=True,
                price_single=100,
                stock_single=6,
                rate_single=1.0,
                sell_stacks=True,
                buy_stacks=True,
                price_stacks=200,
                stock_stacks=8,
                rate_stacks=1.0,
            ),
        }
    )


@pytest.fixture
def manager(populated_fake_db: Database) -> Manager:
    return Manager.from_db(populated_fake_db, name="test")


def test_main(
    populated_fake_db: Database, config: Config, item_list: ItemList, manager: Manager, monkeypatch: MonkeyPatch
) -> None:
    with monkeypatch.context() as m:
        m.setattr(Config, "from_yaml", lambda *a, **kw: config)
        m.setattr(ItemList, "from_csv", lambda *a, **kw: item_list)
        m.setattr(Manager, "create_database_and_manager", lambda *a, **kw: manager)
        ffxiahbot.apps.refill.main(cfg_path=Path("config.yaml"), inp_csvs=[Path("items.csv")], no_prompt=True)

    # check that the items were restocked
    with populated_fake_db.scoped_session() as session:
        for itemid, item in item_list.items.items():
            assert (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 0, AuctionHouse.sale != 0)
                .count()
                == 1
            )
            assert (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 1, AuctionHouse.sale != 0)
                .count()
                == 1
            )
            assert (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 0, AuctionHouse.sale == 0)
                .count()
                == item.stock_single
            )
            assert (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 1, AuctionHouse.sale == 0)
                .count()
                == item.stock_stacks
            )
            assert (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 0, AuctionHouse.sale == 0)
                .first()
                .price
                == item.price_single
            )
            assert (
                session.query(AuctionHouse)
                .filter(AuctionHouse.itemid == itemid, AuctionHouse.stack == 1, AuctionHouse.sale == 0)
                .first()
                .price
                == item.price_stacks
            )



from __future__ import annotations

import datetime
import random
from dataclasses import asdict, dataclass
from string import ascii_letters
from typing import Any

import pandas as pd

from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from ffxiahbot.timeutils import datetime_to_timestamp

RANDOMIZE: object = object()


def randomdt(
    month=None,
    day=None,
    year=None,
    hour=None,
    minute=None,
    second=None,
    microsecond=None,
    tzinfo=datetime.UTC,
    month_range=(1, 12),
    day_range=(1, 31),
    year_range=(1970, 2000),
    hour_range=(0, 23),
    minute_range=(0, 59),
    second_range=(0, 59),
    microsecond_range=(0, 0),
):
    """
    Create a random datetime object.
    """
    if month is None:
        month = random.randint(*month_range)

    if day is None:
        day = random.randint(*day_range)

    if year is None:
        year = random.randint(*year_range)

    if hour is None:
        hour = random.randint(*hour_range)

    if minute is None:
        minute = random.randint(*minute_range)

    if second is None:
        second = random.randint(*second_range)

    if microsecond is None:
        microsecond = random.randint(*microsecond_range)

    for i in range(3):
        try:
            return datetime.datetime(year, month, day - i, hour, minute, second, microsecond, tzinfo)
        except ValueError:
            pass

    return datetime.datetime(year, month, day - 3, hour, minute, second, microsecond, tzinfo)


@dataclass
class AuctionHouseRowBuilder:
    """
    A class to build a row for the AuctionHouse table.
    """

    itemid: int | None | type[RANDOMIZE] = RANDOMIZE
    stack: int | None | type[RANDOMIZE] = RANDOMIZE
    seller: int | None | type[RANDOMIZE] = RANDOMIZE
    seller_name: str | None | type[RANDOMIZE] = RANDOMIZE
    date: int | None | type[RANDOMIZE] = RANDOMIZE
    price: int | None | type[RANDOMIZE] = RANDOMIZE
    buyer_name: str | None | type[RANDOMIZE] = None
    sale: int | None | type[RANDOMIZE] = 0
    sell_date: int | None | type[RANDOMIZE] = None

    @classmethod
    def many(cls, count: int, **kwargs: Any) -> tuple[AuctionHouseRowBuilder, ...]:
        return tuple(cls(**kwargs) for _ in range(count))

    def __post_init__(self):
        self.itemid = self._make_random_int(self.itemid, 0, 2**16 - 1)
        self.stack = self._make_random_int(self.stack, 0, 1)
        self.seller = self._make_random_int(self.seller, 0, 2**32 - 1)
        self.seller_name = self._make_random_str(self.seller_name, 15)
        self.date = self._make_random_date(self.date)
        self.price = self._make_random_int(self.price, 0, 2**32 - 1)
        self.buyer_name = self._make_random_str(self.buyer_name, 15)
        self.sale = self._make_random_int(self.sale, 0, 2**32 - 1)
        self.sell_date = self._make_random_date(self.sell_date)

    def make_orm_object(self) -> AuctionHouse:  # noqa: C901
        def _():
            if self.itemid is not None:
                yield "itemid", AuctionHouse.validate_itemid(self.itemid)
            if self.stack is not None:
                yield "stack", AuctionHouse.validate_stack(self.stack)
            if self.seller is not None:
                yield "seller", AuctionHouse.validate_seller(self.seller)
            if self.seller_name is not None:
                yield "seller_name", AuctionHouse.validate_seller_name(self.seller_name)
            if self.date is not None:
                yield "date", AuctionHouse.validate_date(self.date)
            if self.price is not None:
                yield "price", AuctionHouse.validate_price(self.price)
            if self.buyer_name is not None:
                yield "buyer_name", AuctionHouse.validate_buyer_name(self.buyer_name)
            if self.sale is not None:
                yield "sale", AuctionHouse.validate_sale(self.sale)
            if self.sell_date is not None:
                yield "sell_date", AuctionHouse.validate_sell_date(self.sell_date)

        return AuctionHouse(**dict(_()))

    def add_item_for_sale(self, db: Database) -> None:
        with db.scoped_session() as session:
            session.add(self.make_orm_object())

    @classmethod
    def add_items_for_sale(cls, db: Database, *items: AuctionHouseRowBuilder) -> None:
        for item in items:
            item.add_item_for_sale(db)

    def validate_query_result(self, row: AuctionHouse, **kwargs: Any) -> None:
        if "id" in kwargs:
            assert row.id == kwargs.pop("id")

        assert row.itemid == kwargs.pop("itemid", self.itemid)
        assert row.stack == kwargs.pop("stack", self.stack)
        assert row.seller == kwargs.pop("seller", self.seller)
        assert row.seller_name == kwargs.pop("seller_name", self.seller_name)
        assert row.date == kwargs.pop("date", self.date)
        assert row.price == kwargs.pop("price", self.price)
        assert row.buyer_name == kwargs.pop("buyer_name", self.buyer_name)
        assert row.sale == kwargs.pop("sale", self.sale if self.sale is not None else 0)
        assert row.sell_date == kwargs.pop("sell_date", self.sell_date if self.sell_date is not None else 0)

    @staticmethod
    def _make_random_int(value: int | None | type[RANDOMIZE], lo: int, hi: int) -> int | None:
        return value if value is not RANDOMIZE else random.randint(lo, hi)

    @staticmethod
    def _make_random_str(value: str | None | type[RANDOMIZE], size: int) -> str | None:
        return value if value is not RANDOMIZE else str("".join(random.choices(ascii_letters, k=size)))

    @staticmethod
    def _make_random_date(value: int | None | type[RANDOMIZE]) -> int | None:
        return value if value is not RANDOMIZE else datetime_to_timestamp(randomdt())


DTYPES = {
    "itemid": "uint32",
    "stack": "uint8",
    "seller": "uint32",
    "seller_name": "string",
    "date": "uint32",
    "price": "uint32",
    "buyer_name": "string",
    "sale": "UInt32",
    "sell_date": "UInt32",
}


def setup_ah_transactions(db: Database, *transactions: AuctionHouseRowBuilder) -> pd.DataFrame:
    AuctionHouseRowBuilder.add_items_for_sale(db, *transactions)
    frame = pd.DataFrame([asdict(t) for t in transactions]).astype(DTYPES)
    frame["builder"] = transactions
    return frame


#
# @dataclass
# class Seller:
#     seller: int
#     seller_name: str
#     sold_stacks: dict[int, int] = field(default_factory=dict)
#     sold_singles: dict[int, int] = field(default_factory=dict)
#     selling_stacks: dict[int, int] = field(default_factory=dict)
#     selling_singles: dict[int, int] = field(default_factory=dict)
#     sold_stacks_builders_objects: dict[int, tuple[AuctionHouseRowBuilder, ...]] = field(default_factory=dict)
#     sold_singles_builders_objects: dict[int, tuple[AuctionHouseRowBuilder, ...]] = field(default_factory=dict)
#     selling_stacks_builders_objects: dict[int, tuple[AuctionHouseRowBuilder, ...]] = field(default_factory=dict)
#     selling_singles_builders_objects: dict[int, tuple[AuctionHouseRowBuilder, ...]] = field(default_factory=dict)
#
#
# @dataclass
# class ItemInfo:
#     total: int
#     min_price: int
#     max_price: int
#
#
# @dataclass
# class ItemInfos:
#     ids: set[int]
#     sold: dict[int, ItemInfo]
#     selling: dict[int, ItemInfo]
#
#
# def setup_ah_sales(db: Database, *sellers: Seller) -> ItemInfos:
#     known_itemids: set[int] = set()
#     sold_stacks_total: Counter[int] = Counter()
#     sold_singles_total: Counter[int] = Counter()
#     selling_stacks_total: Counter[int] = Counter()
#     selling_singles_total: Counter[int] = Counter()
#
#     for seller in sellers:
#         # put of stacks for sale
#         for itemid, count in seller.selling_stacks.items():
#             seller.selling_stacks_builders_objects[itemid] = AuctionHouseRowBuilder.many(
#                 count,
#                 db=db,
#                 seller=seller.seller,
#                 seller_name=seller.seller_name,
#                 itemid=itemid,
#                 stack=1,
#             )
#             known_itemids.add(itemid)
#             selling_stacks_total[itemid] += count
#
#         # put of singles for sale
#         for itemid, count in seller.selling_singles.items():
#             seller.selling_singles_builders_objects[itemid] = AuctionHouseRowBuilder.many(
#                 count,
#                 db=db,
#                 seller=seller.seller,
#                 seller_name=seller.seller_name,
#                 itemid=itemid,
#                 stack=0,
#             )
#             known_itemids.add(itemid)
#             selling_singles_total[itemid] += count
#
#         # add history of stacks sold
#         for itemid, count in seller.sold_stacks.items():
#             seller.sold_stacks_builders_objects[itemid] = AuctionHouseRowBuilder.many(
#                 count,
#                 db=db,
#                 seller=seller.seller,
#                 seller_name=seller.seller_name,
#                 itemid=itemid,
#                 stack=1,
#                 buyer_name=RANDOMIZE,
#                 sale=RANDOMIZE,
#                 sell_date=RANDOMIZE,
#             )
#             known_itemids.add(itemid)
#             sold_stacks_total[itemid] += count
#
#         # add history of singles sold
#         for itemid, count in seller.sold_singles.items():
#             seller.sold_singles_builders_objects[itemid] = AuctionHouseRowBuilder.many(
#                 count,
#                 db=db,
#                 seller=seller.seller,
#                 seller_name=seller.seller_name,
#                 itemid=itemid,
#                 stack=0,
#                 buyer_name=RANDOMIZE,
#                 sale=RANDOMIZE,
#                 sell_date=RANDOMIZE,
#             )
#             known_itemids.add(itemid)
#             sold_singles_total[itemid] += count
#
#         return known_itemids, sold_stacks_total, sold_singles_total, selling_stacks_total, selling_singles_total



import contextlib
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any

import sqlalchemy.orm
from sqlalchemy import text
from sqlalchemy.exc import SQLAlchemyError

from ffxiahbot.database import Database


@dataclass(frozen=True)
class Worker:
    """
    Base class for Auction House objects.

    Args:
        db: The database object.
    """

    #: The database object.
    db: Database
    #: Should the session fail on error?
    fail: bool
    #: Should the session rollback on error?
    rollback: bool

    def session(self, **kwargs: Any) -> sqlalchemy.orm.Session:
        """
        Create database session.
        """
        return self.db.session(**kwargs)

    @contextlib.contextmanager
    def scoped_session(self, **kwargs: Any) -> Iterator:
        """
        Create scoped database session.
        """
        _kwargs = {"rollback": self.rollback, "fail": self.fail}
        _kwargs.update(**kwargs)
        try:
            with self.db.scoped_session(**_kwargs) as session:
                yield session
        finally:
            pass

    def can_connect(self) -> bool:
        """
        Test database connection.
        """
        try:
            with self.scoped_session(fail=True) as session:
                session.execute(text("SELECT 1"))
        except SQLAlchemyError:
            return False
        return True



from __future__ import annotations

import contextlib
import re
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any, cast

from bs4 import Tag

from ffxiahbot.logutils import logger
from ffxiahbot.scrubbing.enums import ServerID
from ffxiahbot.scrubbing.scrubber import Scrubber

REGEX_CATEGORY: re.Pattern = re.compile(r"/browse/(\d+)/?.*")
REGEX_ITEM: re.Pattern = re.compile(r"/item/(\d+)")
REGEX_NAME: re.Pattern = re.compile(r"(.*?)\s*-?\s*(FFXIAH)?\.(com)?")

ItemIDsType = list[int] | set[int] | tuple[int, ...]
CatURLsType = list[str] | set[str] | tuple[str, ...]


@dataclass()
class FFXIAHScrubber(Scrubber):
    """
    Get item data from the ffxiah.com website.
    """

    #: The FFXI server
    server: ServerID = ServerID.BAHAMUT
    #: number of threads to use
    num_threads: int | None = None

    def scrub(
        self,
        cat_urls: CatURLsType | None = None,
        item_ids: ItemIDsType | None = None,
    ) -> tuple[dict, dict]:
        """
        Get item metadata main function.

        The item ids can be loaded from category urls or simply passed as a list.  The urls
        can be generated automatically, in which case all possible items will be downloaded.

        Args:
            cat_urls: A preset list of category URLs.
            item_ids: A preset list of item IDs.

        Returns:
            failed: failed item ids
            data: item data
        """
        if not item_ids:
            if not cat_urls:
                cat_urls = sorted(self._get_category_urls(), key=lambda x: list(map(float, re.findall(r"\d+", x))))

            logger.debug("# urls = %d", len(cat_urls))
            item_ids = sorted(set(self._get_itemids(cat_urls)))
        else:
            logger.debug("using passed ids")
            item_ids = sorted(set(item_ids))

            if cat_urls:
                logger.warning("passed urls ignored")

        failed, data = self._get_item_data(item_ids)
        logger.debug("item count = %d", len(item_ids))
        logger.debug("data count = %d", len(data))

        return failed, data

    # step 1
    def _get_category_urls(self) -> Generator[str]:
        """
        Parse http://www.ffxiah.com/browse to get URLs of the form http://www.ffxiah.com/{CategoryNumber}.
        """
        logger.debug("getting category urls")

        # the browse section of FFXIAH has a list of urls with category numbers
        path = "http://www.ffxiah.com/browse"
        logger.debug("open %s", path)
        soup = self.soup(path)
        for tag in soup.find_all("a"):
            if tag.has_attr("href"):
                href = tag.get("href")
                match = REGEX_CATEGORY.match(href)
                if match:
                    try:
                        category = int(match.group(1))
                        if category < 240:
                            yield f"http://www.ffxiah.com{href}"
                            logger.debug("category %s", href)
                        else:
                            logger.debug("skipping %s", href)

                    except (ValueError, IndexError):
                        logger.exception("failed to extract category")
                else:
                    logger.debug("ignoring %s", href)

    # step 2
    def _get_itemids(self, urls: CatURLsType) -> Generator[int]:
        """
        Scrub urls of the form http://www.ffxiah.com/{CategoryNumber} for itemids.
        """
        logger.info("getting itemids")
        for i, url in enumerate(urls):
            logger.info("category %02d/%02d : %s", i + 1, len(urls), url)
            yield from self._get_itemids_for_category_url(url)

    # step 2.1
    def _get_itemids_for_category_url(self, url: str) -> Generator[int]:
        """
        Scrub url of the form http://www.ffxiah.com/{CategoryNumber} for itemids.
        """
        # create tag soup
        soup = self.soup(url)

        # look for table class
        table = soup.find("table", class_="stdlist")
        if not table:
            logger.error("failed to parse <table>")
            return

        # look for table body
        tbody = table.find("tbody")
        if not tbody:
            logger.error("failed to parse <tbody>")
            return

        # look for table rows
        trs = cast(Tag, tbody).find_all("tr")
        if not trs:
            logger.error("failed to parse <tr>")
            return

        # loop table rows
        count = 0
        for j, row in enumerate(trs):
            # look for 'a' tag
            a = row.find("a")

            if a is not None:
                # look for href attr
                href = a.get("href")

                if href is not None:
                    # make sure href matches /item/{number}
                    if (match := REGEX_ITEM.match(href)) is None:
                        logger.error("failed to extract itemid!\n\n\trow %d of %s\n\n%s\n\n", j, url, row)
                    else:
                        item = int(match.group(1))
                        count += 1
                        yield item
                else:
                    logger.error("failed to extract href!\n\n\trow %d of %s\n\n%s\n\n", j, url, row)
            else:
                logger.error("failed to extract 'a' tag!\n\n\trow %d of %s\n\n%s\n\n", j, url, row)

        # make sure we found something
        if not count:
            logger.error("could not find any itemids!")

    # step 3
    def _get_item_data(self, itemids: list[int]) -> tuple[dict[int, Exception], dict[int, dict[str, Any]]]:
        """
        Get metadata for many items.
        """
        logger.info("getting data")

        data = {}
        failed = {}

        # get data from itemids
        for i, itemid in enumerate(itemids):
            try:
                result = self._get_item_data_for_itemid(itemid, index=i, total=len(itemids))
                data[itemid] = result
            except Exception as e:
                logger.exception("failed to scrub %d!", itemid)
                failed[itemid] = e

        if failed:
            for itemid in failed:
                logger.error("failed to scrub %d!", itemid)

        return failed, data

    # step 3.1
    def _get_item_data_for_itemid(self, itemid: int, index: int = 0, total: int = 0) -> dict[str, Any]:
        """
        Get metadata for single item.
        """
        percent = float(index) / float(total) * 100.0 if total > 0 else 0.0

        data: dict[str, Any] = {"name": None, "itemid": itemid}
        url = self._create_item_url(itemid)

        # create tag soup
        logger.debug("open server=%s (%06d/%06d,%6.2f) %s", self.server, index, total, percent, url)
        soup = self.soup(url, absolute=True, sid=self.server.value)

        # extract name
        if (
            soup.title is not None
            and soup.title.text is not None
            and (name_match := REGEX_NAME.match(soup.title.text)) is not None
        ):
            data.update(name=name_match.group(1))
        else:
            data.update(name=None)

        # extract numbers
        for tag in soup.find_all("span", "number-format"):
            try:
                key = tag.parent.find_previous_sibling().text.lower()
                data[key] = int(float(tag.text))
            except (AttributeError, ValueError):
                pass

        # extract rate
        for tag in soup.find_all("span", "sales-rate"):
            with contextlib.suppress(AttributeError, ValueError):
                data["rate"] = float(tag.text)

        # fix key
        data = self._fix_stack_price_key(data)

        return data

    def _get_item_data_for_itemid_map(self, args: tuple[int, ...]) -> dict[str, Any]:
        return self._get_item_data_for_itemid(*args)

    # step 3.1.1
    @staticmethod
    def _create_item_url(itemid: int) -> str:
        """
        Create URL from itemid.
        """
        return f"http://www.ffxiah.com/item/{itemid}"

    # step 3.1.2
    @staticmethod
    def _fix_stack_price_key(data: dict[str, Any]) -> dict[str, Any]:
        """
        Fix dictionary key.
        """
        new_key = r"stack price"

        for key in list(data.keys()):
            if "stack" in key.lower():
                data[new_key] = data[key]

        return data


def extract(data: dict[int, dict[str, Any]], itemid: int, **kwargs: Any) -> dict:
    """
    Extract item data from scrubbed info.
    """
    # singles
    try:
        price_single, sell_single = data[itemid]["median"], True

        # do not sell items without a price
        if price_single <= 0:
            price_single, sell_single = None, False

    except KeyError:
        price_single, sell_single = None, False

    # stacks
    try:
        price_stacks, sell_stacks = data[itemid]["stack price"], True

        # do not sell items without a price
        if price_stacks <= 0:
            price_stacks, sell_stacks = None, False

    except KeyError:
        price_stacks, sell_stacks = None, False

    # the name doesn't really matter
    try:
        name = data[itemid]["name"]
    except KeyError:
        name = None

    result = {
        "name": name,
        "price_single": price_single,
        "stock_single": 5,
        "sell_single": sell_single,
        "buy_single": True,
        "rate_single": 1.0,
        "price_stacks": price_stacks,
        "stock_stacks": 5,
        "sell_stacks": sell_stacks,
        "buy_stacks": True,
        "rate_stacks": 1.0,
    }
    result.update(**kwargs)

    return result



import contextlib
import logging
import warnings
from collections.abc import Iterator
from typing import Any

logger = logging.getLogger("ffxiahbot")


# noinspection PyUnusedLocal
def custom_warning_format(
    message: Warning | str, category: type[Warning], filename: str, lineno: int, *args: Any, **kwargs: Any
) -> str:
    return f"{filename}:{lineno}\n{category.__name__}: {message}"


@contextlib.contextmanager
def capture(capture_warnings: bool = True, fail: bool = False) -> Iterator:
    """
    Log exceptions and warnings.
    """
    default_warning_format = warnings.formatwarning
    try:
        if capture_warnings:
            warnings.formatwarning = custom_warning_format
            logging.captureWarnings(True)
        try:
            yield
        except Exception as e:
            logging.exception("caught unhandled exception")
            if fail and not isinstance(e, Warning):
                raise RuntimeError("application failure") from None
    finally:
        if capture_warnings:
            warnings.formatwarning = default_warning_format
            logging.captureWarnings(False)



from dataclasses import dataclass

from ffxiahbot.auction.worker import Worker
from ffxiahbot.logutils import capture, logger
from ffxiahbot.tables.auctionhouse import AuctionHouse


@dataclass(frozen=True)
class Cleaner(Worker):
    """
    Auction House cleaner.
    """

    def clear(self, seller: int | None = None) -> None:
        """
        Clear out auction house.

        Args:
            seller: The seller to clear out (if None, all rows are cleared).
        """
        # clear rows
        if seller is None:
            # perform query
            with self.scoped_session() as session:
                n = session.query(AuctionHouse).delete()
                logger.info("%d rows dropped", n)

        # clear rows of seller
        else:
            # validate seller
            with capture(fail=self.fail):
                if not isinstance(seller, int) or not seller >= 0:
                    raise RuntimeError("invalid seller: %s", seller)

                # perform query
                with self.scoped_session() as session:
                    n = (
                        session.query(AuctionHouse)
                        .filter(
                            AuctionHouse.seller == seller,
                        )
                        .delete()
                    )
                    logger.info("%d rows dropped", n)

    def count(self, seller: int | None = None) -> int:
        """
        Count the number of rows that would be dropped.

        Args:
            seller: The seller to count (if None, all rows are counted).

        Returns:
            The number of rows.
        """
        # count rows
        if seller is None:
            # perform query
            with self.scoped_session() as session:
                n = session.query(AuctionHouse).count()
                return int(n)

        # count rows of seller
        else:
            # validate seller
            with capture(fail=self.fail):
                if not isinstance(seller, int) or not seller >= 0:
                    raise RuntimeError("invalid seller: %s", seller)

                # perform query
                with self.scoped_session() as session:
                    n = (
                        session.query(AuctionHouse)
                        .filter(
                            AuctionHouse.seller == seller,
                        )
                        .count()
                    )
                    return int(n)



import contextlib
import os
import tempfile
import unittest
from pathlib import Path
from typing import Any, ClassVar

from ffxiahbot import REPOSITORY_BIN_DIR
from ffxiahbot.item import Item
from ffxiahbot.itemlist import ItemList


class TestCase01(unittest.TestCase):
    def setUp(self):
        self.item_list = ItemList()

    def test_init(self):
        pass

    def test_add1(self):
        self.item_list.add(0)
        self.item_list.add(1)

    def test_add2(self):
        self.item_list.add(0)
        with self.assertRaises(KeyError):
            self.item_list.add(0)

    def test_set(self):
        self.item_list.add(0)
        self.item_list.add(1)
        self.item_list.set(0, 1, price_single=5)
        self.assertEqual(self.item_list[0].price_single, 5)
        self.assertEqual(self.item_list[1].price_single, 5)

    def test_getitem(self):
        i0 = self.item_list.add(0)
        self.assertEqual(id(i0), id(self.item_list[0]))

    def test_len(self):
        self.item_list.add(0)
        self.item_list.add(1)
        self.assertEqual(len(self.item_list), 2)

    _test_item1: ClassVar[dict[str, Any]] = {
        "itemid": 0,
        "name": "A",
        "sell_single": True,
        "sell_stacks": False,
        "buy_single": False,
        "buy_stacks": True,
        "price_single": 10,
        "price_stacks": 20,
        "stock_single": 30,
        "stock_stacks": 40,
        "rate_single": 0.5,
        "rate_stacks": 0.6,
        "buy_rate_single": 0.7,
        "buy_rate_stacks": 0.8,
    }

    def test_save_csv(self):
        self.item_list.add(**self._test_item1)
        i, csv_path = tempfile.mkstemp()
        self.item_list.save_csv(Path(csv_path))
        with open(csv_path) as handle:
            handle.readline().strip()
            handle.readline().strip()
        with contextlib.suppress(OSError):
            os.remove(csv_path)

    def test_load_csv(self):
        item_list_1 = ItemList()
        item_list_1.add(**self._test_item1)
        i, csv_path = tempfile.mkstemp()
        item_list_1.save_csv(Path(csv_path))

        item_list_2 = ItemList()
        item_list_2.load_csv(Path(csv_path))

        for k in Item.model_fields:
            attr1 = getattr(item_list_1.get(0), k)
            attr2 = getattr(item_list_2.get(0), k)
            self.assertEqual(attr1, attr2)

    @staticmethod
    def _get_item_list(text, item_list):
        i, csv_path = tempfile.mkstemp()
        with open(csv_path, "w") as handle:
            handle.write(text)

        item_list.load_csv(csv_path)
        with contextlib.suppress(OSError):
            os.remove(csv_path)

    def test_load_csv2(self):
        text = """
itemid, name # comment 0
     0,    A
     2,    B
     4,    C # comment 1
itemid, name, price_single, rate_single
     6,    D,      10,    1.0
"""[1:-1]
        self._get_item_list(text, self.item_list)
        self.assertEqual(self.item_list[0].name, "A")
        self.assertEqual(self.item_list[2].name, "B")
        self.assertEqual(self.item_list[4].name, "C")
        self.assertEqual(self.item_list[6].name, "D")
        self.assertEqual(self.item_list[0].price_single, 1)
        self.assertEqual(self.item_list[2].price_single, 1)
        self.assertEqual(self.item_list[4].price_single, 1)
        self.assertEqual(self.item_list[6].price_single, 10)

    def test_load_csv3(self):
        text = """
itemid, price_single
     0,    -1.0
"""[1:-1]
        with self.assertRaises(ValueError):
            self._get_item_list(text, self.item_list)

    def test_load_csv4(self):
        text1 = """
itemid, name # comment 0
     0,    A
"""[1:-1]

        text2 = """
itemid, name # comment 0
     1,    B
"""[1:-1]

        item_list = ItemList()

        i, csv_path = tempfile.mkstemp()
        with open(csv_path, "w") as handle:
            handle.write(text1)

        item_list.load_csv(csv_path)
        with contextlib.suppress(OSError):
            os.remove(csv_path)

        i, csv_path = tempfile.mkstemp()
        with open(csv_path, "w") as handle:
            handle.write(text2)

        item_list.load_csv(csv_path)
        with contextlib.suppress(OSError):
            os.remove(csv_path)

        self.assertTrue(len(item_list), 2)


def test_load_default_items_csv() -> None:
    if (csv_path := REPOSITORY_BIN_DIR / "items.csv").exists():
        item_list = ItemList()
        item_list.load_csv(csv_path)



from __future__ import annotations

import contextlib
import logging
from collections.abc import Generator
from typing import Any

import sqlalchemy.exc
import sqlalchemy.orm
from sqlalchemy import URL


class Database:
    """
    Database connection using sqlalchemy.

    Args:
        url: sql database connection url
    """

    def __init__(self, url: str | URL, **kwargs: Any) -> None:
        super().__init__()

        # connect
        self.engine = sqlalchemy.create_engine(url, **kwargs)

        # create Session object
        self._Session = sqlalchemy.orm.sessionmaker(bind=self.engine)

    def session(self, **kwargs: Any) -> sqlalchemy.orm.Session:
        """
        Create session.
        """
        return self._Session(**kwargs)

    @contextlib.contextmanager
    def scoped_session(self, rollback: bool = True, fail: bool = False) -> Generator[sqlalchemy.orm.Session]:
        """
        Provide a transactional scope around a series of operations.

        Args:
            rollback: rollback transactions after catch
            fail: raise error after catch
        """
        session = self._Session()
        try:
            yield session

            # commit transactions
            session.commit()

        # catch errors
        except sqlalchemy.exc.SQLAlchemyError as e:
            # log the error
            logging.exception("caught SQL exception")

            # rollback transactions
            if rollback:
                session.rollback()

            # reraise error
            if fail:
                raise e from None

        # cleanup
        finally:
            session.close()

    @classmethod
    def sqlite(cls, database: str = ":memory:", **kwargs: Any) -> Database:
        """
        Alternate constructor.  dialect=sqlite (in-memory)
        """
        url = URL.create("sqlite", database=database)
        return cls(url, **kwargs)

    @classmethod
    def pymysql(cls, hostname: str, database: str, username: str, password: str, port: int, **kwargs: Any) -> Database:
        """
        Alternate constructor.  dialect=mysql, driver=pymysql

        Args:
            hostname: database hostname
            database: database name
            username: database username
            password: database password
            port: database port
        """
        url = URL.create(
            "mysql+pymysql",
            username=username,
            password=password,  # plain (unescaped) text
            host=hostname,
            database=database,
            port=port,
        )
        return cls(url, **kwargs)

    def __str__(self) -> str:
        return repr(self.engine)



import logging
import time
from dataclasses import dataclass
from typing import Any

import bs4
import requests
from bs4 import BeautifulSoup

TIMEOUT: int = 1024


@dataclass()
class Scrubber:
    # noinspection PyBroadException
    @staticmethod
    def soup(url: str, absolute: bool = False, **kwargs: Any) -> BeautifulSoup:
        """
        Open URL and create tag soup.

        :param url: website string
        :type url: str

        :param absolute: perform double get request to find absolute url
        :type absolute: bool
        """
        handle = ""
        max_tries = 10
        for i in range(max_tries):
            # noinspection PyPep8
            try:
                if absolute:
                    url = requests.get(url, timeout=TIMEOUT).url
                handle = requests.get(url, params=kwargs, timeout=TIMEOUT).text
                break
            except Exception:
                logging.exception("urlopen failed (attempt %d)", i + 1)
                if i == max_tries - 1:
                    logging.exception("the maximum urlopen attempts have been reached")
                    raise
                time.sleep(1)

        try:
            s = BeautifulSoup(handle, features="html5lib")
        except bs4.FeatureNotFound:
            s = BeautifulSoup(handle, features="html.parser")

        return s



import os
import unittest

from ffxiahbot import common


class TestCase01(unittest.TestCase):
    def test_backup1(self):
        path = common.backup("common.pz", copy=False)
        self.assertIsNone(path)

    def test_backup2(self):
        path = common.backup(os.path.join(os.path.dirname(__file__), "test_common.py"), copy=False)
        self.assertEqual(os.path.dirname(path), os.path.dirname(__file__))
        self.assertEqual(os.path.basename(path), "test_common.py.1")



from subprocess import run

from mkdocs_macros.plugin import MacrosPlugin


def define_env(env: MacrosPlugin) -> None:
    @env.macro
    def get_help_message(*args: str):
        command = ["uv", "run", "python", "-m", *args, "--help"]
        message = run(command, capture_output=True).stdout.decode("utf-8")
        return f"```bash\n➜  {' '.join(command)}\n{message}\n```"



import pytest

from ffxiahbot.auction.cleaner import Cleaner
from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from tests.cookbook import RANDOMIZE as R
from tests.cookbook import AuctionHouseRowBuilder as AHR
from tests.cookbook import setup_ah_transactions


@pytest.fixture()
def transactions() -> tuple[AHR, ...]:
    return (
        *AHR.many(count=1, price=1, seller=1, seller_name="A", itemid=1, stack=1, buyer_name="X", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=2, seller_name="B", itemid=2, stack=1, buyer_name="X", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=1, seller_name="A", itemid=3, stack=1),
        *AHR.many(count=1, price=1, seller=2, seller_name="B", itemid=4, stack=1),
    )


def test_clear_all(populated_fake_db: Database, transactions: tuple[AHR, ...]) -> None:
    setup_ah_transactions(populated_fake_db, *transactions)
    # ensure database has 4 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 4

    # clear all rows
    cleaner = Cleaner(populated_fake_db, rollback=True, fail=True)
    cleaner.clear(seller=None)

    # ensure database has 0 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 0


def test_clear_for_seller(populated_fake_db: Database, transactions: tuple[AHR, ...]) -> None:
    setup_ah_transactions(populated_fake_db, *transactions)
    # ensure database has 4 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 4

    # clear rows for seller 1
    cleaner = Cleaner(populated_fake_db, rollback=True, fail=True)
    cleaner.clear(seller=1)

    # ensure database has 2 rows
    with populated_fake_db.scoped_session() as session:
        assert session.query(AuctionHouse).count() == 2
        assert session.query(AuctionHouse).filter(AuctionHouse.seller == 1).count() == 0
        assert session.query(AuctionHouse).filter(AuctionHouse.seller == 2).count() == 2



"""
Clear the auction house.
"""

from pathlib import Path
from typing import Annotated

from typer import Option, confirm

from ffxiahbot.common import OptionalPath
from ffxiahbot.config import Config
from ffxiahbot.database import Database
from ffxiahbot.logutils import logger
from ffxiahbot.tables.base import Base


def main(
    cfg_path: Annotated[Path, Option("--config", help="Config file path.")] = Path("config.yaml"),
    no_prompt: Annotated[bool, Option("--no-prompt", help="Do not ask for confirmation.")] = False,
    clear_all: Annotated[bool, Option("--all", help="Clear all items.")] = False,
    use_sqlite_db: Annotated[OptionalPath, Option(help="Use a test SQLite database instead of the real one?")] = None,
) -> None:
    """
    Delete items from the auction house ([red]dangerous operation![/]).
    """
    from ffxiahbot.auction.manager import Manager

    config: Config = Config.from_yaml(cfg_path)
    logger.info("%s", config.model_dump_json(indent=2))

    # create auction house manager
    if use_sqlite_db:
        manager = Manager.from_db(
            db=Database.sqlite(database=str(use_sqlite_db)),
            name=config.name,
            fail=config.fail,
        )
        Base.metadata.create_all(manager.db.engine)
    else:
        manager = Manager.create_database_and_manager(
            hostname=config.hostname,
            database=config.database,
            username=config.username,
            password=config.password,
            port=config.port,
            name=config.name,
            fail=config.fail,
        )

    # clear all items
    if clear_all:
        # really?
        if total := manager.cleaner.count(seller=None):
            if no_prompt or confirm(f"Clear all items? ({total} rows will be removed)", abort=True, show_default=True):
                manager.cleaner.clear(seller=None)
        else:
            logger.info("No items to clear.")
    # clear seller items
    else:
        if total := manager.cleaner.count(seller=manager.seller.seller):
            if no_prompt or confirm(
                f"Clear {config.name}'s items? ({total} rows will be removed)", abort=True, show_default=True
            ):
                manager.cleaner.clear(seller=manager.seller.seller)
        else:
            logger.info("No items to clear.")



import datetime
import unittest

from ffxiahbot import timeutils
from tests.cookbook import randomdt


class BaseTest(unittest.TestCase):
    N = 10000

    def assertInRange(self, n, a, b):
        self.assertGreaterEqual(n, a)
        self.assertLessEqual(n, b)


class TestCase01(BaseTest):
    def test_construct(self):
        for _i in range(self.N):
            randomdt()

    def test_range(self):
        r = (1, 2)
        for name in ("month", "day", "year", "hour", "minute", "second", "microsecond"):
            for _i in range(10):
                n = getattr(randomdt(**{f"{name}_range": r}), name)
                self.assertInRange(n, *r)

    def test_explicit(self):
        v = 1
        for name in ("month", "day", "year", "hour", "minute", "second", "microsecond"):
            n = getattr(randomdt(**{f"{name}": v}), name)
            self.assertEqual(n, v)


class TestCase02(BaseTest):
    def test_str_to_datetime(self):
        dobj = timeutils.str_to_datetime("1/2/1971 04:05:06")
        self.assertTrue(isinstance(dobj, datetime.datetime))
        self.assertEqual(dobj.month, 1)
        self.assertEqual(dobj.day, 2)
        self.assertEqual(dobj.year, 1971)
        self.assertEqual(dobj.hour, 4)
        self.assertEqual(dobj.minute, 5)
        self.assertEqual(dobj.second, 6)
        self.assertEqual(dobj.microsecond, 0)

    def test_datetime_to_str(self):
        dobj = datetime.datetime(1971, 1, 2, 4, 5, 6, 0)
        sobj = timeutils.datetime_to_str(dobj)
        self.assertTrue(isinstance(sobj, str))
        self.assertEqual(sobj, "01/02/1971 04:05:06")


class TestCase03(BaseTest):
    def test_datetime_timestamp_conversion(self):
        for _i in range(self.N):
            dobj1 = randomdt(year_range=(1971, 2015))
            stmp1 = timeutils.datetime_to_timestamp(dobj1)
            dobj2 = timeutils.timestamp_to_datetime(stmp1)
            stmp2 = timeutils.datetime_to_timestamp(dobj2)
            self.assertEqual(dobj1, dobj2)
            self.assertEqual(stmp1, stmp2)


class TestCase04(BaseTest):
    def test_str(self):
        dobj = timeutils.datetime("1/2/1903 04:05:06")
        self.assertTrue(isinstance(dobj, datetime.datetime))
        self.assertEqual(dobj.month, 1)
        self.assertEqual(dobj.day, 2)
        self.assertEqual(dobj.year, 1903)
        self.assertEqual(dobj.hour, 4)
        self.assertEqual(dobj.minute, 5)
        self.assertEqual(dobj.second, 6)
        self.assertEqual(dobj.microsecond, 0)

    def test_timestamp(self):
        for _i in range(self.N):
            dobj1 = randomdt(year_range=(1971, 2015))
            stmp1 = timeutils.datetime_to_timestamp(dobj1)
            dobj2 = timeutils.datetime(stmp1)
            self.assertTrue(isinstance(dobj1, datetime.datetime))
            self.assertEqual(dobj1, dobj2)

    def test_datetime(self):
        for _i in range(self.N):
            dobj1 = randomdt()
            dobj2 = timeutils.datetime(dobj1)
            self.assertTrue(isinstance(dobj2, datetime.datetime))
            self.assertEqual(dobj1, dobj2)

    def test_args(self):
        dobj = timeutils.datetime(1900, 1, 2)
        self.assertTrue(isinstance(dobj, datetime.datetime))
        self.assertEqual(dobj.month, 1)
        self.assertEqual(dobj.day, 2)
        self.assertEqual(dobj.year, 1900)

    def test_kwargs(self):
        dobj = timeutils.datetime(1900, 1, 2, microsecond=1)
        self.assertTrue(isinstance(dobj, datetime.datetime))
        self.assertEqual(dobj.month, 1)
        self.assertEqual(dobj.day, 2)
        self.assertEqual(dobj.year, 1900)
        self.assertEqual(dobj.microsecond, 1)


class TestCase05(BaseTest):
    def test_str(self):
        stmp1 = timeutils.timestamp("1/2/1971 04:05:06")
        self.assertTrue(isinstance(stmp1, int))
        dobj1 = timeutils.timestamp_to_datetime(stmp1)
        self.assertEqual(dobj1.month, 1)
        self.assertEqual(dobj1.day, 2)
        self.assertEqual(dobj1.year, 1971)
        self.assertEqual(dobj1.hour, 4)
        self.assertEqual(dobj1.minute, 5)
        self.assertEqual(dobj1.second, 6)
        self.assertEqual(dobj1.microsecond, 0)

    def test_timestamp(self):
        for _i in range(self.N):
            dobj1 = randomdt(year_range=(1971, 2015))
            stmp1 = timeutils.datetime_to_timestamp(dobj1)
            stmp2 = timeutils.timestamp(stmp1)
            self.assertTrue(isinstance(stmp1, int))
            self.assertEqual(stmp1, stmp2)

    def test_datetime(self):
        for _i in range(self.N):
            dobj1 = randomdt()
            stmp1 = timeutils.timestamp(dobj1)
            stmp2 = timeutils.datetime_to_timestamp(dobj1)
            self.assertTrue(isinstance(stmp1, int))
            self.assertEqual(stmp1, stmp2)

    def test_args(self):
        stmp1 = timeutils.timestamp(1971, 1, 2, tzinfo=datetime.UTC)
        dobj1 = timeutils.timestamp_to_datetime(stmp1)
        self.assertTrue(isinstance(dobj1, datetime.datetime))
        self.assertEqual(dobj1.month, 1)
        self.assertEqual(dobj1.day, 2)
        self.assertEqual(dobj1.year, 1971)

    def test_kwargs(self):
        stmp1 = timeutils.timestamp(1971, 1, 2, tzinfo=datetime.UTC)
        dobj1 = timeutils.timestamp_to_datetime(stmp1)
        self.assertTrue(isinstance(dobj1, datetime.datetime))
        self.assertEqual(dobj1.month, 1)
        self.assertEqual(dobj1.day, 2)
        self.assertEqual(dobj1.year, 1971)



from datetime import UTC, datetime

import pytest
from sqlalchemy.orm import Session

from ffxiahbot.auction.buyer import Buyer
from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from ffxiahbot.timeutils import datetime_to_timestamp
from tests.cookbook import RANDOMIZE as R
from tests.cookbook import AuctionHouseRowBuilder as AHR
from tests.cookbook import setup_ah_transactions


@pytest.fixture()
def buy_date() -> datetime:
    return datetime(2000, 1, 1, tzinfo=UTC)


@pytest.fixture()
def buy_timestamp(buy_date: datetime) -> int:
    return datetime_to_timestamp(buy_date)


@pytest.fixture()
def transactions() -> tuple[AHR, ...]:
    return (
        *AHR.many(count=1, price=1, seller=1, seller_name="A", itemid=1, stack=1, buyer_name="W", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=2, seller_name="B", itemid=2, stack=1, buyer_name="X", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=3, seller_name="C", itemid=3, stack=0, buyer_name="Y", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=4, seller_name="D", itemid=4, stack=0, buyer_name="Z", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=5, seller_name="E", itemid=5, stack=1),
        *AHR.many(count=1, price=1, seller=6, seller_name="F", itemid=6, stack=1),
        *AHR.many(count=1, price=1, seller=7, seller_name="G", itemid=7, stack=0),
        *AHR.many(count=1, price=1, seller=8, seller_name="H", itemid=8, stack=0),
    )


# noinspection PyTypeChecker
def get_row_for_sale_by(session: Session, seller_name: str) -> AuctionHouse:
    return session.query(AuctionHouse).filter(AuctionHouse.seller_name == seller_name).first()


def test_buyer_buys_row(
    populated_fake_db: Database,
    transactions: tuple[AHR, ...],
    buy_date: datetime,
    buy_timestamp: int,
    seller_name: str = "E",
    buyer_name: str = "X",
    buy_price: int = 1,
) -> None:
    expected_transaction_table = setup_ah_transactions(populated_fake_db, *transactions)

    buyer = Buyer(populated_fake_db, buyer_name=buyer_name, rollback=True, fail=True)
    with populated_fake_db.scoped_session() as session:
        row = get_row_for_sale_by(session, seller_name)
        buyer.set_row_buyer_info(row, 0, buy_price)

    with populated_fake_db.scoped_session() as session:
        for _, expected_transaction in expected_transaction_table.iterrows():
            if expected_transaction.seller_name != seller_name:
                row = get_row_for_sale_by(session, expected_transaction.seller_name)
            else:
                row = get_row_for_sale_by(session, seller_name)
                expected_transaction.builder.sale = buy_price
                expected_transaction.builder.buyer_name = buyer_name
                expected_transaction.builder.validate_query_result(row)

            expected_transaction.builder.validate_query_result(row)


def test_buyer_raises_on_invalid_item(
    populated_fake_db: Database,
    transactions: tuple[AHR, ...],
    buy_date: datetime,
    buy_timestamp: int,
    seller_name: str = "A",
    buyer_name: str = "X",
    buy_price: int = 1,
):
    setup_ah_transactions(populated_fake_db, *transactions)
    buyer = Buyer(populated_fake_db, buyer_name=buyer_name, rollback=True, fail=True)
    with populated_fake_db.scoped_session() as session:
        row = get_row_for_sale_by(session, seller_name)
        with pytest.raises(RuntimeError):
            buyer.set_row_buyer_info(row, buy_timestamp, buy_price)



from typing import Any

import pytest

from ffxiahbot.item import Item


@pytest.mark.parametrize(
    "kwargs,error",
    [
        pytest.param({"itemid": 0, "name": "A"}, None, id="name=A"),
        pytest.param({"itemid": 0, "price_single": 1}, None, id="price_single=1"),
        pytest.param({"itemid": 0, "price_single": 0}, ValueError, id="price_single=0 raises"),
        pytest.param({"itemid": 0, "price_single": -1}, ValueError, id="price_single=-1 raises"),
        pytest.param({"itemid": 0, "price_stacks": 1}, None, id="price_stacks=1"),
        pytest.param({"itemid": 0, "price_stacks": 0}, ValueError, id="price_stacks=0 raises"),
        pytest.param({"itemid": 0, "price_stacks": -1}, ValueError, id="price_stacks=-1 raises"),
        pytest.param({"itemid": 0, "stock_single": 0}, None, id="stock_single=0"),
        pytest.param({"itemid": 0, "stock_single": -1}, ValueError, id="stock_single=-1"),
        pytest.param({"itemid": 0, "stock_stacks": 0}, None, id="stock_stacks=0"),
        pytest.param({"itemid": 0, "stock_stacks": -1}, ValueError, id="stock_stacks=-1"),
        pytest.param({"itemid": 0, "sell_rate_single": 0.0}, None, id="sell_rate_single=0.0"),
        pytest.param({"itemid": 0, "sell_rate_single": 0.5}, None, id="sell_rate_single=0.5"),
        pytest.param({"itemid": 0, "sell_rate_single": 1.0}, None, id="sell_rate_single=1.0"),
        pytest.param({"itemid": 0, "sell_rate_single": -1.5}, ValueError, id="sell_rate_single=-1.5"),
        pytest.param({"itemid": 0, "sell_rate_single": +1.5}, ValueError, id="sell_rate_single=+1.5"),
        pytest.param({"itemid": 0, "sell_rate_stacks": 0.0}, None, id="sell_rate_stacks=0.0"),
        pytest.param({"itemid": 0, "sell_rate_stacks": 0.5}, None, id="sell_rate_stacks=0.5"),
        pytest.param({"itemid": 0, "sell_rate_stacks": 1.0}, None, id="sell_rate_stacks=1.0"),
        pytest.param({"itemid": 0, "sell_rate_stacks": -1.5}, ValueError, id="sell_rate_stacks=-1.5"),
        pytest.param({"itemid": 0, "sell_rate_stacks": +1.5}, ValueError, id="sell_rate_stacks=+1.5"),
        pytest.param({"itemid": 0, "buy_rate_single": 0.0}, None, id="buy_rate_single=0.0"),
        pytest.param({"itemid": 0, "buy_rate_single": 0.5}, None, id="buy_rate_single=0.5"),
        pytest.param({"itemid": 0, "buy_rate_single": 1.0}, None, id="buy_rate_single=1.0"),
        pytest.param({"itemid": 0, "buy_rate_single": -1.5}, ValueError, id="buy_rate_single=-1.5"),
        pytest.param({"itemid": 0, "buy_rate_single": +1.5}, ValueError, id="buy_rate_single=+1.5"),
        pytest.param({"itemid": 0, "buy_rate_stacks": 0.0}, None, id="buy_rate_stacks=0.0"),
        pytest.param({"itemid": 0, "buy_rate_stacks": 0.5}, None, id="buy_rate_stacks=0.5"),
        pytest.param({"itemid": 0, "buy_rate_stacks": 1.0}, None, id="buy_rate_stacks=1.0"),
        pytest.param({"itemid": 0, "buy_rate_stacks": -1.5}, ValueError, id="buy_rate_stacks=-1.5"),
        pytest.param({"itemid": 0, "buy_rate_stacks": +1.5}, ValueError, id="buy_rate_stacks=+1.5"),
    ],
)
def test_create_item_raises(kwargs: dict[str, Any], error: type[Exception] | None):
    if error:
        with pytest.raises(error):
            Item(**kwargs)
    else:
        item = Item(**kwargs)
        for key, value in kwargs.items():
            assert getattr(item, key) == value



"""
Buy and sell items on the auction house.
"""

from datetime import datetime
from pathlib import Path
from typing import Annotated, Any

from apscheduler.schedulers.blocking import BlockingScheduler
from typer import Option

from ffxiahbot.common import OptionalPath, OptionalPathList
from ffxiahbot.config import Config
from ffxiahbot.database import Database
from ffxiahbot.logutils import logger
from ffxiahbot.tables.base import Base


def main(
    cfg_path: Annotated[Path, Option("--config", help="Config file path.")] = Path("config.yaml"),
    inp_csvs: Annotated[
        OptionalPathList, Option("--inp-csv", help="Input CSV file path.", show_default="items.csv")
    ] = None,
    buy_items: Annotated[bool, Option(help="Enable the buying of items.")] = True,
    sell_items: Annotated[bool, Option(help="Enable the selling of items.")] = True,
    buy_immediately: Annotated[
        bool, Option("--buy-immediately", help="Buy items immediately instead of waiting?")
    ] = False,
    restock_immediately: Annotated[
        bool, Option("--restock-immediately", help="Restock items immediately instead of waiting?")
    ] = False,
    use_sqlite_db: Annotated[OptionalPath, Option(help="Use a test SQLite database instead of the real one?")] = None,
) -> None:
    """
    Run a bot that buys and sells items on the auction house continuously.
    """
    from ffxiahbot.auction.manager import Manager
    from ffxiahbot.itemlist import ItemList

    if inp_csvs is None:
        inp_csvs = [Path("items.csv")]

    config: Config = Config.from_yaml(cfg_path)
    logger.info("%s", config.model_dump_json(indent=2))

    if not buy_items and not sell_items:
        raise RuntimeError("both buying and selling are disabled! nothing to do...")

    # create auction house manager
    if use_sqlite_db:
        manager = Manager.from_db(
            db=Database.sqlite(database=str(use_sqlite_db)),
            name=config.name,
            fail=config.fail,
        )
        Base.metadata.create_all(manager.db.engine)
    else:
        manager = Manager.create_database_and_manager(
            hostname=config.hostname,
            database=config.database,
            username=config.username,
            password=config.password,
            port=config.port,
            name=config.name,
            fail=config.fail,
        )

    # test database connection
    if not manager.can_connect():
        raise RuntimeError("cannot connect to the database!")

    # load data
    item_list = ItemList.from_csv(*inp_csvs)

    # run callbacks on a schedule
    scheduler = BlockingScheduler()

    if buy_items:
        buy_kwargs: dict[str, Any] = {} if not buy_immediately else {"next_run_time": datetime.now().astimezone()}
        scheduler.add_job(
            lambda: manager.buy_items(item_list=item_list, use_buying_rates=config.use_buying_rates),
            trigger="interval",
            id="buy_items",
            seconds=config.tick,
            max_instances=1,
            name="Buy Items",
            **buy_kwargs,
        )

    if sell_items:
        sell_kwargs: dict[str, Any] = {} if not restock_immediately else {"next_run_time": datetime.now().astimezone()}
        scheduler.add_job(
            lambda: manager.restock_items(item_list=item_list, use_selling_rates=config.use_selling_rates),
            trigger="interval",
            id="restock_items",
            seconds=config.restock,
            max_instances=1,
            next_run_time=None if not restock_immediately else datetime.now().astimezone(),
            name="Restock Items",
            **sell_kwargs,
        )

    scheduler.start()



from pathlib import Path

import pytest
from pytest import MonkeyPatch

import ffxiahbot.apps.clear
from ffxiahbot.auction.manager import Manager
from ffxiahbot.config import Config
from ffxiahbot.database import Database
from ffxiahbot.tables.auctionhouse import AuctionHouse
from tests.cookbook import RANDOMIZE as R
from tests.cookbook import AuctionHouseRowBuilder as AHR
from tests.cookbook import setup_ah_transactions


@pytest.fixture
def config() -> Config:
    return Config(name="A")


@pytest.fixture
def manager(populated_fake_db: Database) -> Manager:
    return Manager.from_db(populated_fake_db, name="A")


@pytest.fixture()
def transactions() -> tuple[AHR, ...]:
    return (
        *AHR.many(count=1, price=1, seller=0, seller_name="A", itemid=1, stack=1, buyer_name="X", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=2, seller_name="B", itemid=2, stack=1, buyer_name="X", sell_date=R, sale=R),
        *AHR.many(count=1, price=1, seller=0, seller_name="A", itemid=3, stack=1),
        *AHR.many(count=1, price=1, seller=2, seller_name="B", itemid=4, stack=1),
    )


@pytest.mark.parametrize("clear_all", [True, False])
def test_main(
    populated_fake_db: Database,
    config: Config,
    manager: Manager,
    transactions: tuple[AHR, ...],
    clear_all: bool,
    monkeypatch: MonkeyPatch,
) -> None:
    setup_ah_transactions(populated_fake_db, *transactions)

    with monkeypatch.context() as m:
        m.setattr(Config, "from_yaml", lambda *a, **kw: config)
        m.setattr(Manager, "create_database_and_manager", lambda *a, **kw: manager)

        # ensure database has rows
        with populated_fake_db.scoped_session() as session:
            assert session.query(AuctionHouse).count() == len(transactions)

        ffxiahbot.apps.clear.main(cfg_path=Path("config.yaml"), no_prompt=True, clear_all=clear_all)

        # ensure database was cleared
        with populated_fake_db.scoped_session() as session:
            if clear_all:
                assert session.query(AuctionHouse).count() == 0
            else:
                assert session.query(AuctionHouse).count() == sum(1 for t in transactions if t.seller != 0)



"""
Create item database.
"""

from pathlib import Path
from typing import Annotated

from click import Choice
from typer import Exit, Option

from ffxiahbot.common import OptionalIntList, OptionalStrList
from ffxiahbot.logutils import logger
from ffxiahbot.scrubbing.enums import ServerID

# noinspection PyProtectedMember
ServerIDOption = Option(
    "--server",
    help="The server name (asura, bahamut, etc).",
    click_type=Choice(ServerID._member_names_, case_sensitive=False),
    metavar="SERVER",
)


def main(
    cfg_path: Annotated[Path, Option("--config", help="Config file path.")] = Path("config.yaml"),
    out_csv: Annotated[Path, Option("--out-csv", help="The output CSV file to save.")] = Path("items.csv"),
    server_str: Annotated[str, ServerIDOption] = ServerID.ASURA.name,
    cat_urls: Annotated[OptionalStrList, Option("--cat-url", help="Preset category URLs.")] = None,
    item_ids: Annotated[OptionalIntList, Option("--item-id", help="Preset item IDs.")] = None,
    overwrite: Annotated[bool, Option("--overwrite", help="Overwrite output CSV?")] = False,
    stock_single: Annotated[int, Option(help="The default number of items for singles.")] = 10,
    stock_stacks: Annotated[int, Option(help="The default number of items for stacks.")] = 10,
    should_backup: Annotated[bool, Option("--backup", help="Backup output CSV?")] = False,
) -> None:
    """
    Download a list of item prices from ffxiah.com and save to a CSV file.
    """
    from ffxiahbot.common import backup
    from ffxiahbot.config import Config
    from ffxiahbot.itemlist import ItemList
    from ffxiahbot.scrubbing.ffxiah import FFXIAHScrubber, extract

    config: Config = Config.from_yaml(cfg_path)
    logger.info("%s", config.model_dump_json(indent=2))

    # check output file name validity
    if not out_csv.suffix.lower() == ".csv":
        raise ValueError("--out-csv file must be a CSV file!")

    if not overwrite and not should_backup and out_csv.exists():
        logger.error("output file already exists!\n\t%s", out_csv)
        logger.error("please use --overwrite or --backup")
        raise Exit(-1)

    # scrub data
    scrubber = FFXIAHScrubber(server=ServerID[server_str.upper()])
    failed, data = scrubber.scrub(cat_urls=cat_urls, item_ids=item_ids)

    if data:
        # create item list from data
        item_list = ItemList()
        for itemid in sorted(data.keys()):
            kwargs = extract(data, itemid, stock_single=stock_single, stock_stacks=stock_stacks)
            item_list.add(itemid, **kwargs)

        # backup file
        if should_backup:
            backup(out_csv, copy=True)

        # overwrites if exists, but we checked already
        item_list.save_csv(out_csv)
    else:
        raise RuntimeError("no items were scrubbed!")

    if failed:
        raise RuntimeError("not all item ids were scrubbed, but a CSV was still saved!")



from __future__ import annotations

import ast
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, cast

from ffxiahbot.item import Item, allowed_item_keys, item_csv_title_str, item_csv_value_str
from ffxiahbot.logutils import logger


@dataclass
class ItemList:
    """
    Container for Item objects.
    """

    items: dict[int, Item] = field(default_factory=dict)

    @classmethod
    def from_csv(cls, *csv_paths: Path) -> ItemList:
        """
        Create ItemList from CSV file(s).

        Args:
            *csv_paths: The path(s) to the CSV file(s).

        Returns:
            ItemList: The ItemList object created from the CSV file(s).
        """
        # make sure there is data
        if not csv_paths:
            raise RuntimeError("missing item data CSV!")

        # load data
        item_list = cls()
        logger.info("loading item data...")
        for csv_path in csv_paths:
            item_list.load_csv(csv_path)

        return item_list

    def add(self, itemid: int, **kwargs: Any) -> Item:
        """
        Add Item to ItemList.  Item must not already exist.

        Args:
            itemid: The item id.
            **kwargs: The keyword arguments to pass to the Item constructor.

        Returns:
            Item: The Item object created.
        """
        item = Item(itemid=itemid, **kwargs)
        if item.itemid in self.items:
            raise KeyError("duplicate item found: %d" % item.itemid)
        self.items[item.itemid] = item
        return item

    def set(self, *itemids: int, **kwargs: Any) -> None:
        """
        Set Item(s) properties.

        Args:
            *itemids: The item id(s) to set the properties for.
            **kwargs: The properties to set.

        Raises:
            KeyError: If a property is not found.
        """
        for itemid in itemids:
            i = self[itemid]
            for k in kwargs:
                if hasattr(i, k):
                    setattr(i, k, kwargs[k])
                else:
                    raise KeyError(f"{k!s}")

    def get(self, itemid: int) -> Item:
        """
        Get Item by itemid.

        Args:
            itemid: The item id.

        Returns:
            Item: The Item object.
        """
        return self.items[itemid]

    def __getitem__(self, itemid: int) -> Item:
        """
        Get Item by itemid.

        Args:
            itemid: The item id.

        Returns:
            Item: The Item object.
        """
        return self.items[itemid]

    def __contains__(self, itemid: int) -> bool:
        """
        Check if Item is in ItemList.

        Args:
            itemid: The item id.

        Returns:
            bool: True if the Item is in the ItemList, False otherwise.
        """
        return itemid in self.items

    def __len__(self) -> int:
        """
        Get the number of items in the ItemList.

        Returns:
            int: The number of items in the ItemList.
        """
        return len(self.items)

    def load_csv(self, csv_path: Path | str) -> None:  # noqa: C901
        """
        Load Item(s) from CSV file.

        Columns are Item attributes.  The 'itemid' column is required.

        Args:
            csv_path: The path to the CSV file.

        Raises:
            RuntimeError: If the itemid column is missing.
            RuntimeError: If an unknown column is found.
            RuntimeError: If something is wrong with a line.
        """
        regex_c = re.compile(r"#.*$")
        regex_t = re.compile("[{0}{1}]?True[{0}{1}]?".format('"', "'"), re.IGNORECASE)
        regex_f = re.compile("[{0}{1}]?False[{0}{1}]?".format('"', "'"), re.IGNORECASE)

        logger.info("load %s", csv_path)
        line_number = 0
        with open(csv_path) as handle:
            # first line is item titles
            line = handle.readline()
            line_number += 1

            # ignore comments
            line = regex_c.sub("", line).strip()

            # split into tokens
            keys = line.split(",")
            keys = [x.strip().lower() for x in keys]

            # make sure keys are valid
            for k in keys:
                if k not in Item.model_fields:
                    raise RuntimeError(f"unknown column: {k}")

            # check for primary key
            if "itemid" not in keys:
                raise RuntimeError(f"missing itemid column:\n\t{keys}")

            # other lines are items
            line = handle.readline()
            line_number += 1

            while line:
                # remove comments
                line = regex_c.sub("", line).strip()

                # fix True and False
                line = regex_t.sub("1", line)
                line = regex_f.sub("0", line)

                # ignore empty lines
                if line:
                    # split into tokens
                    tokens: list[str | int | None] = [x.strip() for x in line.split(",")]

                    # check for new title line
                    if set(tokens).issubset(allowed_item_keys()):
                        keys = cast(list[str], tokens)

                        # check for primary key
                        if "itemid" not in keys:
                            raise RuntimeError(f"missing itemid column:\n\t{keys}")

                    # validate line
                    elif set(tokens).intersection(allowed_item_keys()):
                        raise RuntimeError("something wrong with line %d" % line_number)

                    # process normal line
                    else:
                        # try to evaluate tokens
                        for i, token in enumerate(tokens):
                            if isinstance(token, str):
                                try:
                                    token = ast.literal_eval(token)
                                except SyntaxError:
                                    pass
                                except ValueError:
                                    pass
                                except NameError:
                                    pass

                            # process missing tokens
                            if isinstance(token, str) and not token:
                                token = None

                            tokens[i] = token

                        # map values
                        kwargs: dict[str, str | int | None] = {k: None for k in keys}
                        for i in range(len(tokens)):
                            kwargs[keys[i]] = tokens[i]

                        # add new item
                        itemid: int = cast(int, kwargs.pop("itemid"))
                        self.add(itemid, **kwargs)

                # read next line
                line = handle.readline()
                line_number += 1

    def save_csv(self, csv_path: Path, iter_title: int = 1024) -> None:
        """
        Save Item data to CSV file.

        Args:
            csv_path: The path to save the CSV file.
            iter_title: The number of items to write before writing the title row.
        """
        if csv_path.exists():
            logger.info("overwriting file...")
            logger.info("save %s", csv_path)
        else:
            logger.info("save %s", csv_path)

        with csv_path.open("w") as stream:
            for j, i in enumerate(self.items):
                if j % iter_title == 0:
                    stream.write(item_csv_title_str())
                stream.write(item_csv_value_str(self.items[i]))



from dataclasses import dataclass

from ffxiahbot.auction.worker import Worker
from ffxiahbot.logutils import logger
from ffxiahbot.tables.auctionhouse import AuctionHouse


@dataclass(frozen=True)
class Buyer(Worker):
    """
    Auction House buyer.

    Args:
        buyer_name: The name of the character buying the item.
    """

    #: The name of the character buying the item.
    buyer_name: str

    def set_row_buyer_info(self, row: AuctionHouse, date: int, price: int) -> None:
        """
        Update the AuctionHouse row with buyer information.
        This does not actually buy the item unless a database commit is made.

        Args:
            row: The AuctionHouse row.
            date: The date of the sale.
            price: The price of the sale.
        """
        # validate
        if row.sale != 0:
            raise RuntimeError("item not for sale!")

        row.buyer_name = AuctionHouse.validate_buyer_name(self.buyer_name)
        row.sell_date = AuctionHouse.validate_date(date)
        row.sale = AuctionHouse.validate_price(price)
        logger.info("%s", row)



from enum import IntEnum


# noinspection SpellCheckingInspection
class ServerID(IntEnum):
    BAHAMUT: int = 1
    SHIVA: int = 2
    TITAN: int = 3
    RAMUH: int = 4
    PHOENIX: int = 5
    CARBUNCLE: int = 6
    FENRIR: int = 7
    SYLPH: int = 8
    VALEFOR: int = 9
    ALEXANDER: int = 10
    LEVIATHAN: int = 11
    ODIN: int = 12
    IFRIT: int = 13
    DIABOLOS: int = 14
    CAITSITH: int = 15
    QUETZALCOATL: int = 16
    SIREN: int = 17
    UNICORN: int = 18
    GILGAMESH: int = 19
    RAGNAROK: int = 20
    PANDEMONIUM: int = 21
    GARUDA: int = 22
    CERBERUS: int = 23
    KUJATA: int = 24
    BISMARCK: int = 25
    SERAPH: int = 26
    LAKSHMI: int = 27
    ASURA: int = 28
    MIDGARDSORMR: int = 29
    FAIRY: int = 30
    REMORA: int = 31
    HADES: int = 32



from sqlalchemy import text
from sqlalchemy.dialects.mysql import INTEGER, SMALLINT, TINYINT, VARCHAR
from sqlalchemy.orm import Mapped, mapped_column

from ffxiahbot import timeutils
from ffxiahbot.tables.base import Base

_template = """
[AuctionHouseRow]
    addr        = {addr}
    id          = {self.id}
    itemid      = {self.itemid}
    stack       = {self.stack}
    seller      = {self.seller}
    seller_name = {self.seller_name}
    date        = {self.date}
    price       = {self.price}
    buyer_name  = {self.buyer_name}
    sale        = {self.sale}
    sell_date   = {self.sell_datestr}
"""[:-1]


class AuctionHouse(Base):
    __tablename__ = "auction_house"

    id: Mapped[int] = mapped_column(INTEGER(10, unsigned=True), nullable=False, autoincrement=True, primary_key=True)
    itemid: Mapped[int] = mapped_column(
        SMALLINT(5, unsigned=True), nullable=False, index=True, server_default=text("'0'")
    )
    stack: Mapped[int] = mapped_column(TINYINT(1, unsigned=True), nullable=False, server_default=text("'0'"))
    seller: Mapped[int] = mapped_column(
        INTEGER(10, unsigned=True), nullable=False, index=True, server_default=text("'0'")
    )
    seller_name: Mapped[str] = mapped_column(VARCHAR(15), nullable=True)
    date: Mapped[int] = mapped_column(INTEGER(10, unsigned=True), nullable=False, server_default=text("'0'"))
    price: Mapped[int] = mapped_column(INTEGER(10, unsigned=True), nullable=False, server_default=text("'0'"))
    buyer_name: Mapped[str] = mapped_column(VARCHAR(15), nullable=True)
    sale: Mapped[int] = mapped_column(INTEGER(10, unsigned=True), nullable=False, server_default=text("'0'"))
    sell_date: Mapped[int] = mapped_column(INTEGER(10, unsigned=True), nullable=False, server_default=text("'0'"))

    def __repr__(self) -> str:
        return f"({hex(id(self))}) AuctionHouseRow id={self.id}"

    def __str__(self) -> str:
        return _template.format(self=self, addr=hex(id(self)))

    @staticmethod
    def validate_itemid(itemid: int) -> int:
        itemid = int(itemid)
        if itemid < 0:
            raise ValueError(f"itemid={itemid} is negative")
        return itemid

    @staticmethod
    def validate_stack(stack: int | bool) -> int:
        if stack:
            return 1
        return 0

    @staticmethod
    def validate_seller(seller: int) -> int:
        seller = int(seller)
        if seller < 0:
            raise ValueError(f"seller={seller} is negative")
        return seller

    @staticmethod
    def validate_seller_name(seller_name: str) -> str:
        if len(seller_name) > 15:
            raise ValueError(f"seller_name={seller_name} is too long")
        return seller_name

    @staticmethod
    def validate_date(date: int) -> int:
        stamp = timeutils.timestamp(date)
        if (dt := timeutils.timestamp_to_datetime(stamp)).year < 1970:
            raise ValueError(f"date={dt} is before 1970")
        return stamp

    @staticmethod
    def validate_price(price: int) -> int:
        price = int(price)
        if price < 0:
            raise ValueError(f"price={price} is negative")
        return price

    @staticmethod
    def validate_sale(sale: int) -> int:
        sale = int(sale)
        if sale < 0:
            raise ValueError(f"sale={sale} is negative")
        return sale

    @staticmethod
    def validate_buyer_name(buyer_name: str) -> str:
        if len(buyer_name) > 15:
            raise ValueError(f"buyer_name={buyer_name} is too long")
        return buyer_name

    @staticmethod
    def validate_sell_date(sell_date: int) -> int:
        stamp = timeutils.timestamp(sell_date)
        if (dt := timeutils.timestamp_to_datetime(stamp)).year < 1970:
            raise ValueError(f"sell_date={dt} is before 1970")
        return stamp

    @property
    def sell_datestr(self) -> str:
        if self.sell_date is None:
            return str(None)
        return timeutils.datetime_to_str(timeutils.timestamp_to_datetime(int(self.sell_date)))



"""
Functions for handling dates and times.

.. sourcecode:: ipython

    In [1]: import ffxiahbot.timeutils

The most useful functions are:
    * :py:func:`ffxiahbot.timeutils.timestamp`
    * :py:func:`ffxiahbot.timeutils.datetime`

.. sourcecode:: ipython

    In [2]: ffxiahbot.timeutils.timestamp('01/01/2015 00:00:00')
    Out[2]: 1420070400.0

    In [3]: ffxiahbot.timeutils.datetime('01/01/2015 00:00:00')
    Out[3]: datetime.datetime(2015, 1, 1)

.. seealso:: :py:mod:`datetime`, :py:class:`datetime.datetime`

Classes
-------

.. autosummary::
    :nosignatures:

    ffxiahbot.timeutils.DatetimeToTimestamp

Functions
---------

.. autosummary::
    :nosignatures:

    ffxiahbot.timeutils.str_to_datetime
    ffxiahbot.timeutils.datetime_to_str
    ffxiahbot.timeutils.datetime_to_timestamp
    ffxiahbot.timeutils.timestamp_to_datetime
    ffxiahbot.timeutils.datetime
    ffxiahbot.timeutils.timestamp

Documentation
-------------

.. moduleauthor:: Adam Gagorik <adam.gagorik@gmail.com>
"""

import datetime as _datetime
from typing import Any


def str_to_datetime(date_string: str) -> _datetime.datetime:
    """
    Convert string to datetime object.

    :param date_string: string with format :code:`'%m/%d/%Y %H:%M:%S'`
    :type date_string: str

    :return: datetime object
    :rtype: :py:class:`datetime.datetime`

    .. sourcecode:: ipython

        In [1]: ffxiahbot.timeutils.str_to_datetime('01/01/2015 00:00:00')
        Out[1]: datetime.datetime(2015, 1, 1, 0, 0)

    .. seealso:: :py:meth:`datetime.datetime.strptime`
    """
    return _datetime.datetime.strptime(date_string, "%m/%d/%Y %H:%M:%S").replace(tzinfo=_datetime.timezone.utc)


def datetime_to_str(datetime_obj: _datetime.datetime) -> str:
    """
    Convert datetime object to string.

    :param datetime_obj: datetime object
    :type datetime_obj: :py:class:`datetime.datetime`

    :return: datetime as string
    :rtype: str

    .. sourcecode:: ipython

        In [1]: ffxiahbot.timeutils.datetime_to_str(datetime.datetime(2015, 1, 1))
        Out[1]: '01/01/2015 00:00:00'

    .. seealso:: :py:meth:`datetime.datetime.strftime`
    """
    return datetime_obj.strftime("%m/%d/%Y %H:%M:%S")


class DatetimeToTimestamp:
    """
    Convert datetime object to timestamp.

    Calculates the total seconds since 01/01/1970.

    .. warning:: Don't use this class directly.

    .. seealso:: :py:data:`ffxiahbot.timeutils.datetime_to_timestamp`
    """

    #: 01/01/1970
    epoch: _datetime.datetime = _datetime.datetime(1970, 1, 1, tzinfo=_datetime.timezone.utc)

    def __call__(self, datetime_obj: _datetime.datetime) -> int:
        return int((datetime_obj - self.epoch).total_seconds())


datetime_to_timestamp = DatetimeToTimestamp()
"""
Convert datetime object to timestamp.

:param datetime_obj: datetime object
:type datetime_obj: :py:class:`datetime.datetime`

:return: datetime as integer
:rtype: int

.. sourcecode:: ipython

    In [1]: ffxiahbot.timeutils.datetime_to_timestamp(datetime.datetime(2015, 1, 1, 0, 0))
    Out[1]: 1420070400.0

.. seealso:: :py:meth:`datetime.timedelta.total_seconds`
"""


def timestamp_to_datetime(stamp: int) -> _datetime.datetime:
    """
    Convert timestamp to datetime object.

    :param stamp: seconds since epoch (01/01/1970)
    :type stamp: int, float

    :return: datetime object
    :rtype: :py:class:`datetime.datetime`

    .. sourcecode:: ipython

        In [1]: ffxiahbot.timeutils.timestamp_to_datetime(1420070400)
        Out[1]: datetime.datetime(2015, 1, 1, 0, 0)

    .. seealso:: :py:meth:`datetime.timedelta.utcfromtimestamp`
    """
    return _datetime.datetime.fromtimestamp(stamp, _datetime.timezone.utc)


def datetime(*args: Any, **kwargs: Any) -> _datetime.datetime:
    """
    Convert anything (within reason) to a datetime object.

    When there are multiple arguments:
        * the :py:class:`datetime.datetime` constructor is called

    If there is only one argument:
        * if it is a :py:class:`datetime.datetime` it is returned
        * if it is a :py:obj:`str` it is passed to :py:func:`ffxiahbot.timeutils.str_to_datetime`
        * if it is a :py:obj:`int` it is passed to :py:func:`ffxiahbot.timeutils.timestamp_to_datetime`
        * if it is a :py:obj:`float` it is passed to :py:func:`ffxiahbot.timeutils.timestamp_to_datetime`

    :param args: positional arguments
    :param kwargs: keyword arguments

    .. sourcecode:: ipython

        In [1]: ffxiahbot.timeutils.datetime(datetime.datetime(2015, 1, 1))
        Out[1]: datetime.datetime(2015, 1, 1)

        In [2]: ffxiahbot.timeutils.datetime('01/01/2015 00:00:00')
        Out[2]: datetime.datetime(2015, 1, 1)

        In [3]: ffxiahbot.timeutils.datetime(1420070400.0)
        Out[3]: datetime.datetime(2015, 1, 1)

        In [4]: ffxiahbot.timeutils.datetime(1420070400)
        Out[4]: datetime.datetime(2015, 1, 1)

    .. seealso::

        :py:func:`ffxiahbot.timeutils.str_to_datetime`
        :py:func:`ffxiahbot.timeutils.timestamp_to_datetime`
        :py:class:`datetime.datetime`
    """
    if len(args) > 1 or kwargs:
        return _datetime.datetime(*args, **(kwargs | {"tzinfo": _datetime.timezone.utc}))

    try:
        obj = args[0]
    except IndexError:
        raise ValueError("expecting argument") from None

    if isinstance(obj, _datetime.datetime):
        return obj.replace(tzinfo=_datetime.timezone.utc)

    if isinstance(obj, str):
        return str_to_datetime(obj)

    if isinstance(obj, int):
        return timestamp_to_datetime(obj)

    if isinstance(obj, float):
        return timestamp_to_datetime(int(obj))

    raise TypeError(f"unknown type: {type(obj)}")


def timestamp(*args: Any, **kwargs: Any) -> int:
    """
    Convert anything (within reason) to a timestamp.

    :param args: positional arguments
    :param kwargs: keyword arguments

    * Passes arguments to :py:func:`ffxiahbot.timeutils.datetime`
    * Passes result to :py:data:`ffxiahbot.timeutils.datetime_to_timestamp`

    .. sourcecode:: ipython

        In [1]: ffxiahbot.timeutils.timestamp('01/01/2015 00:00:00')
        Out[1]: 1420070400.0

        In [2]: ffxiahbot.timeutils.timestamp(datetime.datetime(2015, 1, 1))
        Out[2]: 1420070400.0

        In [3]: ffxiahbot.timeutils.timestamp(1420070400.0)
        Out[3]: 1420070400.0

        In [4]: ffxiahbot.timeutils.timestamp(1420070400)
        Out[4]: 1420070400.0

    .. seealso::

        :py:func:`ffxiahbot.timeutils.datetime`
        :py:data:`ffxiahbot.timeutils.datetime_to_timestamp`
    """
    return datetime_to_timestamp(datetime(*args, **kwargs))



from collections.abc import Callable
from dataclasses import dataclass

import sqlalchemy

from ffxiahbot.auction.worker import Worker
from ffxiahbot.logutils import capture
from ffxiahbot.tables.auctionhouse import AuctionHouse


@dataclass(frozen=True)
class Browser(Worker):
    """
    Auction House browser.
    """

    def count(self) -> int:
        """
        Get the number of rows.
        """
        with self.scoped_session() as session:
            return int(session.query(AuctionHouse).count())

    def get_stock(self, itemid: int, stack: bool = False, seller: int | None = None) -> int:
        """
        Get stock of an item.

        Args:
            itemid: The item number.
            stack: Consider stacks?
            seller: Consider seller?

        Returns:
            The total number of singles or stacks being sold, in general or by the specific seller.
        """
        with capture(fail=self.fail), self.scoped_session() as session:
            # ignore seller
            if seller is None:
                n = (
                    session.query(AuctionHouse)
                    .filter(
                        AuctionHouse.itemid == AuctionHouse.validate_itemid(itemid),
                        AuctionHouse.stack == AuctionHouse.validate_stack(stack),
                        AuctionHouse.sale == 0,
                    )
                    .count()
                )
                return int(n)

            # consider seller
            else:
                n = (
                    session.query(AuctionHouse)
                    .filter(
                        AuctionHouse.itemid == AuctionHouse.validate_itemid(itemid),
                        AuctionHouse.seller == AuctionHouse.validate_seller(seller),
                        AuctionHouse.stack == AuctionHouse.validate_stack(stack),
                        AuctionHouse.sale == 0,
                    )
                    .count()
                )
                return int(n)

    def get_price(
        self, itemid: int, stack: bool = False, seller: int | None = None, func: Callable = sqlalchemy.func.min
    ) -> int | None:
        """
        Get the historical price of an item.

        Args:
            itemid: The item number.
            stack: Consider stacks?
            seller: Consider seller?
            func: Aggregation function (min).

        Return:
            The aggregated priced for an item, in general or for the specific seller.
        """
        with capture(fail=self.fail), self.scoped_session() as session:
            # ignore seller
            if seller is None:
                n = (
                    session.query(func(AuctionHouse.sale))
                    .filter(
                        AuctionHouse.itemid == AuctionHouse.validate_itemid(itemid),
                        AuctionHouse.stack == AuctionHouse.validate_stack(stack),
                        AuctionHouse.sale != 0,
                    )
                    .scalar()
                )
                return None if n is None else int(n)

            # consider seller
            else:
                n = (
                    session.query(func(AuctionHouse.sale))
                    .filter(
                        AuctionHouse.itemid == AuctionHouse.validate_itemid(itemid),
                        AuctionHouse.seller == AuctionHouse.validate_seller(seller),
                        AuctionHouse.stack == AuctionHouse.validate_stack(stack),
                        AuctionHouse.sale != 0,
                    )
                    .scalar()
                )
                return None if n is None else int(n)



from __future__ import annotations

import contextlib
import datetime
import random
from collections import Counter
from collections.abc import Generator
from dataclasses import dataclass
from typing import Any

import pandas as pd
from pydantic import SecretStr
from rich.progress import Progress, TaskID, TimeElapsedColumn

from ffxiahbot import timeutils
from ffxiahbot.auction.browser import Browser
from ffxiahbot.auction.buyer import Buyer
from ffxiahbot.auction.cleaner import Cleaner
from ffxiahbot.auction.seller import Seller
from ffxiahbot.auction.worker import Worker
from ffxiahbot.database import Database
from ffxiahbot.itemlist import ItemList
from ffxiahbot.logutils import capture, logger
from ffxiahbot.tables.auctionhouse import AuctionHouse


@dataclass(frozen=True)
class Manager(Worker):
    """
    Auction House browser.
    """

    #: Auction House database row ids to ignore.
    blacklist: set[int]
    #: A worker to browse the auction house.
    browser: Browser
    #: A worker to clean the auction house.
    cleaner: Cleaner
    #: A worker to sell and restock items.
    seller: Seller
    #: A worker to purchase items.
    buyer: Buyer

    @classmethod
    def create_database_and_manager(
        cls,
        hostname: str,
        database: str,
        username: str,
        password: str | SecretStr,
        port: int,
        **kwargs: Any,
    ) -> Manager:
        """
        Create database and manager at the same time.

        Args:
            hostname: Database hostname.
            database: Database name.
            username: Database username.
            password: Database password.
            port: Database port.
            **kwargs: Additional arguments passed to the `Manager.from_db` method.
        """
        # connect to database
        db = Database.pymysql(
            hostname=hostname,
            database=database,
            username=username,
            password=password if isinstance(password, str) else password.get_secret_value(),
            port=port,
        )

        # create auction house manager
        return cls.from_db(db=db, **kwargs)

    @classmethod
    def from_db(
        cls,
        db: Database,
        name: str,
        fail: bool = True,
        rollback: bool = True,
        blacklist: set[int] | None = None,
    ) -> Manager:
        """
        Create manager from database.

        Args:
            db: Database connection.
            name: Name of the seller and buyer.
            fail: If True, raise exceptions.
            rollback: If True, rollback transactions.
            blacklist: Auction House row ids to ignore.
        """
        return cls(
            db=db,
            blacklist=blacklist if blacklist is not None else set(),
            browser=Browser(db=db, rollback=rollback, fail=fail),
            cleaner=Cleaner(db=db, rollback=rollback, fail=fail),
            seller=Seller(db=db, rollback=rollback, fail=fail, seller=0, seller_name=name),
            buyer=Buyer(db=db, rollback=rollback, fail=fail, buyer_name=name),
            rollback=rollback,
            fail=fail,
        )

    def add_to_blacklist(self, rowid: int) -> None:
        """
        Add row to blacklist.

        Args:
            rowid: Row id to blacklist.
        """
        logger.info("blacklisting: row=%d", rowid)
        self.blacklist.add(rowid)

    def buy_items(self, item_list: ItemList, use_buying_rates: bool = False) -> None:  # noqa: C901
        """
        The main buy item loop.

        Args:
            item_list: Item data.
            use_buying_rates: If True, only buy items at their buying rate.
        """
        with self.scoped_session(fail=self.fail) as session:
            # find rows that are still up for sale
            q = session.query(AuctionHouse).filter(
                AuctionHouse.seller != self.seller.seller,
                AuctionHouse.sell_date == 0,
                AuctionHouse.sale == 0,
            )
            # loop rows
            counts: Counter[str] = Counter()
            for row in q:
                if row.id in self.blacklist:
                    logger.debug("skipping blacklisted row %d", row.id)
                    counts["blacklisted"] += 1
                    continue

                if row.itemid not in item_list:
                    logger.error("item missing from database: %d", row.itemid)
                    self.add_to_blacklist(row.id)
                    counts["unknown itemid"] += 1
                    continue

                with capture(fail=self.fail):
                    item = item_list[row.itemid]

                    if row.stack:
                        if not item.buy_stacks:
                            logger.debug("not allowed to buy item! itemid=%d", row.itemid)
                            self.add_to_blacklist(row.id)
                            counts["forbidden item"] += 1
                        else:
                            if not use_buying_rates or random.random() <= item.buy_rate_stacks:
                                if self._buy_row(row, item.price_stacks):
                                    counts["stack purchased"] += 1
                                else:
                                    counts["price too high"] += 1
                            else:
                                counts["buy rate too low"] += 1
                    else:
                        if not item.buy_single:
                            logger.debug("not allowed to buy item! itemid=%d", row.itemid)
                            self.add_to_blacklist(row.id)
                            counts["forbidden item"] += 1
                        else:
                            if not use_buying_rates or random.random() <= item.buy_rate_single:
                                if self._buy_row(row, item.price_single):
                                    counts["single purchased"] += 1
                                else:
                                    counts["price too high"] += 1
                            else:
                                counts["buy rate too low"] += 1

            counts_frame = pd.DataFrame.from_dict(counts, orient="index").rename(columns={0: "count"})
            if not counts_frame.empty:
                logger.debug("manager.buy_items counts: \n%s", counts_frame)
            else:
                logger.warning("no rows processed when buying items (no items for sale?)")

    def _buy_row(self, row: AuctionHouse, max_price: int) -> bool:
        """
        Buy a row.

        Args:
            row: Auction House row to buy.
            max_price: Maximum price to pay.
        """
        # check price
        if row.price <= max_price:
            date = timeutils.timestamp(datetime.datetime.now())
            self.buyer.set_row_buyer_info(row, date, max_price)
            return True
        else:
            logger.info(
                "price too high! itemid=%d %d <= %d",
                row.itemid,
                row.price,
                max_price,
            )
            self.add_to_blacklist(row.id)
            return False

    def restock_items(self, item_list: ItemList, use_selling_rates: bool = False) -> None:
        """
        The main restock loop.

        Args:
            item_list: Item data.
            use_selling_rates: If True, only restock items at their selling rate.
        """
        # loop over items
        with progress_bar("[red]Restocking Items...", total=len(item_list)) as (progress, task):
            for item in item_list.items.values():
                # singles
                if item.sell_single:
                    self._sell_item(
                        item.itemid,
                        stack=False,
                        price=item.price_single,
                        stock=item.stock_single,
                        rate=None if not use_selling_rates else item.sell_rate_single,
                    )
                    progress.update(task, advance=0.5)

                # stacks
                if item.sell_stacks:
                    self._sell_item(
                        item.itemid,
                        stack=True,
                        price=item.price_stacks,
                        stock=item.stock_stacks,
                        rate=None if not use_selling_rates else item.sell_rate_stacks,
                    )
                    progress.update(task, advance=0.5)

    @property
    def _sell_time(self) -> int:
        """
        The timestamp to use for selling items.
        """
        return timeutils.timestamp(datetime.datetime(2099, 1, 1))

    def _sell_item(self, itemid: int, stack: bool, price: int, stock: int, rate: float | None) -> None:
        """
        Sell an item.

        Args:
            itemid: The item id.
            stack: True if selling stacks, False if selling singles.
            price: The price to sell the item for.
            stock: The amount of items to stock.
        """
        # check history
        history_price = self.browser.get_price(itemid=itemid, stack=stack, seller=self.seller.seller)

        # set history
        if history_price is None or history_price <= 0:
            self.seller.set_history(itemid=itemid, stack=stack, price=price, date=self._sell_time, count=1)

        # get stock
        current_stock = self.browser.get_stock(itemid=itemid, stack=stack, seller=self.seller.seller)

        # restock
        if current_stock < stock:
            for _ in range(stock - current_stock):
                if rate is None or random.random() <= rate:
                    self.seller.sell_item(itemid=itemid, stack=stack, date=self._sell_time, price=price, count=1)


@contextlib.contextmanager
def progress_bar(description: str, total: int) -> Generator[tuple[Progress, TaskID]]:
    try:
        with Progress(
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            transient=True,
        ) as progress:
            task = progress.add_task(description=description, total=total)
            yield progress, task
    finally:
        pass



import unittest

from ffxiahbot.scrubbing.ffxiah import FFXIAHScrubber, extract


class TestCase01(unittest.TestCase):
    def setUp(self):
        self.scrubber = FFXIAHScrubber()
        self.scrubber.save = False

    def test_get_category_urls(self):
        self.scrubber._get_category_urls()

    def test_get_itemids_for_category_url(self):
        url = r"http://www.ffxiah.com/browse/49/ninja-tools"
        self.scrubber._get_itemids_for_category_url(url)

    def test_get_itemids(self):
        urls = [
            r"http://www.ffxiah.com/browse/49/ninja-tools",
            r"http://www.ffxiah.com/browse/56/breads-rice",
        ]
        self.scrubber._get_itemids(urls)

    def test_get_item_data_for_itemid(self):
        self.scrubber._get_item_data_for_itemid(4096)

    def test_get_item_data(self):
        self.scrubber._get_item_data(list(range(1, 9)))

    def test_scrub(self):
        self.scrubber.scrub(item_ids={1, 2, 3, 4})

    def test_extract(self):
        extract(self.scrubber._get_item_data([4096])[1], 4096)
